import threading
import time as ttime
import json

from bluesky_queueserver.manager.start_manager import WatchdogProcess
from bluesky_queueserver.tests.common import format_jsonrpc_msg

import logging


class ReManagerEmulation(threading.Thread):
    """
    Emulation of RE Manager, which is using Thread instead of Process.
    The functionality of the emulator is to test if Watchdog can start
    and restart RE Manager properly. The emulator also generates periodic
    'heartbeat' messages to inform RE Manager that it is running.
    """

    def __init__(
        self,
        *args,
        conn_watchdog,
        conn_worker,
        config=None,
        msg_queue=None,
        log_level=logging.DEBUG,
        number_of_restarts,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._conn_watchdog = conn_watchdog
        self.n_loops = 0
        self._exit = False
        self._restart = False
        self._send_heartbeat = True
        self._lock = threading.Lock()
        self._config_dict = config or {}
        self._log_level = log_level
        self._number_of_restarts = number_of_restarts

    def _heartbeat(self):
        hb_period, dt = 0.5, 0.01
        n_wait = round(hb_period / dt)
        msg = format_jsonrpc_msg("heartbeat", {"value": "alive"}, notification=True)
        msg_json = json.dumps(msg)
        while True:
            # Since we are emulating 'kill' method, we want the function to
            #   react to 'exit' quickly.
            for n in range(n_wait):
                ttime.sleep(0.005)
                if self._exit:
                    return
            if self._send_heartbeat:
                with self._lock:
                    self._conn_watchdog.send(msg_json)

    def exit(self, *, restart=False):
        """
        Stop the emulated RE Manager (exit the 'run' method). Set 'restart=True'
        to skip informing Watchdog that the exit is intentional: Watchdog is expected
        to restart the process.
        """
        self._restart = restart
        self._exit = True

    def kill(self):
        """
        This is emulation of 'kill' method of mp.Process. The method just normally
        exists the current process.
        """
        self.exit(restart=True)

    def send_msg_to_watchdog(self, method, params=None, *, notification=False, timeout=0.5):
        # The function may block all communication for the period of 'timeout', but
        #   this is acceptable for testing. Timeout would typically indicate an error.
        msg = format_jsonrpc_msg(method, params, notification=notification)
        with self._lock:
            self._conn_watchdog.send(json.dumps(msg))
            if notification:
                return
            if self._conn_watchdog.poll(timeout):
                response_json = self._conn_watchdog.recv()
                response = json.loads(response_json)
                result = response["result"]
            else:
                result = None
        return result

    def stop_heartbeat(self):
        """
        Heatbeat generator may be stopped to emulate 'freezing' of the event loop of RE Manager.
        """
        self._send_heartbeat = False

    def run(self):
        th_hb = threading.Thread(target=self._heartbeat)
        th_hb.start()

        while not self._exit:
            ttime.sleep(0.01)
            self.n_loops += 1

        if not self._restart:
            msg = format_jsonrpc_msg("manager_stopping", notification=True)
            with self._lock:
                self._conn_watchdog.send(json.dumps(msg))

        th_hb.join()


class ReWorkerEmulation(threading.Thread):
    def __init__(
        self,
        *args,
        conn,
        config=None,
        msg_queue=None,
        log_level=logging.DEBUG,
        user_group_permissions=None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._config_dict = config or {}
        self._exit = False
        self.n_loops = 0
        self._log_level = log_level

    def exit(self):
        self._exit = True

    def kill(self):
        self.exit()

    def run(self):
        while not self._exit:
            ttime.sleep(0.005)
            self.n_loops += 1


def test_WatchdogProcess_1():
    """Test starting and orderly existing the RE Manager"""
    wp = WatchdogProcess(cls_run_engine_manager=ReManagerEmulation)
    wp_th = threading.Thread(target=wp.run)
    wp_th.start()
    ttime.sleep(1)  # Let RE Manager run 1 second
    assert wp._re_manager.n_loops > 0, "RE is not running"
    wp._re_manager.exit(restart=False)
    ttime.sleep(0.05)
    assert wp._manager_is_stopping is True, "'Manager Stopping' flag is not set"
    wp_th.join(0.1)


def test_WatchdogProcess_2():
    """
    Test starting RE Manager, stopping heartbeat generator and waiting for restart of RE Manager.
    Check that the number of restarts of the manager process is properly counted and propagated.
    """
    wp = WatchdogProcess(cls_run_engine_manager=ReManagerEmulation)
    wp_th = threading.Thread(target=wp.run)
    wp_th.start()
    ttime.sleep(1)  # Let RE Manager run 1 second
    assert wp._re_manager.n_loops > 0, "RE is not running"
    assert wp._re_manager_n_restarts == 1

    n_loops = wp._re_manager.n_loops

    wp._re_manager.stop_heartbeat()
    hb_timeout = wp._heartbeat_timeout
    ttime.sleep(hb_timeout + 0.5)
    # At this point RE Manager is expected to run for 0.5 second, so
    #   the new number of loops must be about 'n_loops/2'.
    #   Here we check if RE Manager was really restarted and the number of
    #   loops reset.
    assert wp._re_manager.n_loops < n_loops, "Unexpected number of loops"
    assert wp._re_manager_n_restarts == 2

    wp._re_manager.exit(restart=False)
    ttime.sleep(0.05)
    assert wp._manager_is_stopping is True, "'Manager Stopping' flag is not set"
    wp_th.join(0.1)


def test_WatchdogProcess_3():
    """Test starting RE Manager, exiting without sending notification and
    and waiting for the restart of RE Manager"""
    wp = WatchdogProcess(cls_run_engine_manager=ReManagerEmulation)
    wp_th = threading.Thread(target=wp.run)
    wp_th.start()
    ttime.sleep(1)  # Let RE Manager run 1 second
    assert wp._re_manager.n_loops > 0, "RE is not running"
    n_loops = wp._re_manager.n_loops

    # Stop RE Manager without notifying the Watchdog (emulates crashing of RE Manager)
    wp._re_manager.exit(restart=True)
    hb_timeout = wp._heartbeat_timeout
    ttime.sleep(hb_timeout + 0.5)
    # At this point RE Manager is expected to run for 0.5 second, so
    #   the new number of loops must be about 'n_loops/2'.
    #   Here we check if RE Manager was really restarted and the number of
    #   loops reset.
    assert wp._re_manager.n_loops < n_loops, "Unexpected number of loops"

    wp._re_manager.exit(restart=False)
    ttime.sleep(0.05)
    assert wp._manager_is_stopping is True, "'Manager Stopping' flag is not set"
    wp_th.join(0.1)


def test_WatchdogProcess_4():
    """
    Test if Watchdog correctly executing commands that control starting
    and stopping RE Worker.
    """
    wp = WatchdogProcess(cls_run_engine_manager=ReManagerEmulation, cls_run_engine_worker=ReWorkerEmulation)
    wp_th = threading.Thread(target=wp.run)
    wp_th.start()
    ttime.sleep(0.01)

    response = wp._re_manager.send_msg_to_watchdog("start_re_worker", params={"user_group_permissions": {}})
    assert response["success"] is True, f"Unexpected response from RE Manager: {response}"

    # Worker is expected to be alive
    response = wp._re_manager.send_msg_to_watchdog("is_worker_alive")
    assert response["worker_alive"] is True, f"Unexpected response from RE Manager: {response}"

    # Join running process (thread). Expected to timeout.
    # Note: here timeout should be set to be smaller than timeout for the message
    #   in 'send_msg_to_watchdog method.
    response = wp._re_manager.send_msg_to_watchdog("join_re_worker", {"timeout": 0.1})
    assert response["success"] is False, f"Unexpected response from RE Manager: {response}"

    # Worker is expected to be alive
    response = wp._re_manager.send_msg_to_watchdog("is_worker_alive")
    assert response["worker_alive"] is True, "Unexpected response from RE Manager"

    # Exit the process (thread).
    wp._re_worker.exit()
    ttime.sleep(0.01)

    # Worker is expected to be stopped
    response = wp._re_manager.send_msg_to_watchdog("is_worker_alive")
    assert response["worker_alive"] is False, "Unexpected response from RE Manager"

    response = wp._re_manager.send_msg_to_watchdog("join_re_worker", {"timeout": 0.5})
    assert response["success"] is True, "Unexpected response from RE Manager"

    wp._re_manager.exit(restart=False)
    wp_th.join(0.1)


def test_WatchdogProcess_5():
    """
    Test 'kill_re_worker' command RE Worker.
    """
    wp = WatchdogProcess(cls_run_engine_manager=ReManagerEmulation, cls_run_engine_worker=ReWorkerEmulation)
    wp_th = threading.Thread(target=wp.run)
    wp_th.start()
    ttime.sleep(0.01)

    response = wp._re_manager.send_msg_to_watchdog("start_re_worker", params={"user_group_permissions": {}})
    assert response["success"] is True, "Unexpected response from RE Manager"

    # Worker is expected to be alive
    response = wp._re_manager.send_msg_to_watchdog("is_worker_alive")
    assert response["worker_alive"] is True, "Unexpected response from RE Manager"

    # Kill RE Worker process (emulated, since RE Worker is a thread)
    response = wp._re_manager.send_msg_to_watchdog("kill_re_worker")
    assert response["success"] is True, "Unexpected response from RE Manager"

    # Worker is expected to be stopped
    response = wp._re_manager.send_msg_to_watchdog("is_worker_alive")
    assert response["worker_alive"] is False, "Unexpected response from RE Manager"

    response = wp._re_manager.send_msg_to_watchdog("join_re_worker", {"timeout": 0.5})
    assert response["success"] is True, "Unexpected response from RE Manager"

    wp._re_manager.exit(restart=False)
    wp_th.join(0.1)


def test_WatchdogProcess_6():
    """
    Test if RE configuration is passed to RE Worker
    """
    config_worker = {"some_parameter1": "some_value1"}
    config_manager = {"some_parameter2": "some_value2"}

    wp = WatchdogProcess(
        config_worker=config_worker,
        config_manager=config_manager,
        cls_run_engine_manager=ReManagerEmulation,
        cls_run_engine_worker=ReWorkerEmulation,
    )
    wp_th = threading.Thread(target=wp.run)
    wp_th.start()
    ttime.sleep(0.01)

    response = wp._re_manager.send_msg_to_watchdog("start_re_worker", params={"user_group_permissions": {}})
    assert response["success"] is True, "Unexpected response from RE Manager"

    # Check if configuration was set correctly in RE Worker and RE manager
    assert wp._re_worker._config_dict == config_worker, "Worker configuration was not passed correctly"
    assert wp._re_manager._config_dict == config_manager, "Manager configuration was not passed correctly"

    # Exit the process (thread).
    wp._re_worker.exit()
    ttime.sleep(0.01)

    response = wp._re_manager.send_msg_to_watchdog("join_re_worker", {"timeout": 0.5})
    assert response["success"] is True, "Unexpected response from RE Manager"

    wp._re_manager.exit(restart=False)
    wp_th.join(0.1)


def test_WatchdogProcess_7():
    """
    Test if the Watchdog and Manager processes are initialized with correct logger
    """
    config_worker = {"some_parameter1": "some_value1"}
    config_manager = {"some_parameter2": "some_value2"}

    log_level = logging.INFO

    wp = WatchdogProcess(
        config_worker=config_worker,
        config_manager=config_manager,
        cls_run_engine_manager=ReManagerEmulation,
        cls_run_engine_worker=ReWorkerEmulation,
        log_level=log_level,
    )
    wp_th = threading.Thread(target=wp.run)
    wp_th.start()
    ttime.sleep(0.01)

    response = wp._re_manager.send_msg_to_watchdog("start_re_worker", params={"user_group_permissions": {}})
    assert response["success"] is True, "Unexpected response from RE Manager"

    # Check if configuration was set correctly in RE Worker and RE manager
    assert wp._re_worker._log_level == log_level
    assert wp._re_manager._log_level == log_level

    # Exit the process (thread).
    wp._re_worker.exit()
    ttime.sleep(0.01)

    response = wp._re_manager.send_msg_to_watchdog("join_re_worker", {"timeout": 0.5})
    assert response["success"] is True, "Unexpected response from RE Manager"

    wp._re_manager.exit(restart=False)
    wp_th.join(0.1)



import pytest
import time as ttime
import json
import multiprocessing
import threading
import asyncio
import zmq

from bluesky_queueserver.manager.comms import (
    PipeJsonRpcReceive,
    PipeJsonRpcSendAsync,
    CommTimeoutError,
    CommJsonRpcError,
    ZMQCommSendThreads,
    ZMQCommSendAsync,
    generate_new_zmq_key_pair,
    generate_zmq_public_key,
    validate_zmq_key,
)
from bluesky_queueserver.tests.common import format_jsonrpc_msg


def count_threads_with_name(name):
    """
    Returns the number of currently existing threads with the given name
    """
    n_count = 0
    for th in threading.enumerate():
        if th.name == name:
            n_count += 1
    return n_count


# =======================================================================
#                       Class CommJsonRpcError


def test_CommJsonRpcError_1():
    """
    Basic test for `CommJsonRpcError` class.
    """
    err_msg, err_code, err_type = "Some error occured", 25, "RuntimeError"
    ex = CommJsonRpcError(err_msg, err_code, err_type)

    assert ex.message == err_msg, "Message set incorrectly"
    assert ex.error_code == err_code, "Error code set incorrectly"
    assert ex.error_type == err_type, "Error type set incorrectly"


def test_CommJsonRpcError_2():
    """
    Basic test for `CommJsonRpcError` class.
    """
    err_msg, err_code, err_type = "Some error occured", 25, "RuntimeError"
    ex = CommJsonRpcError(err_msg, err_code, err_type)

    s = str(ex)
    assert err_msg in s, "Error message is not found in printed error message"
    assert str(err_code) in s, "Error code is not found in printed error message"
    assert err_type in s, "Error type is not found in printed error message"

    repr = f"CommJsonRpcError('{err_msg}', {err_code}, '{err_type}')"
    assert ex.__repr__() == repr, "Error representation is printed incorrectly"


def test_CommJsonRpcError_3_fail():
    """
    `CommJsonRpcError` class. Failing cases
    """
    err_msg, err_code, err_type = "Some error occured", 25, "RuntimeError"
    ex = CommJsonRpcError(err_msg, err_code, err_type)

    with pytest.raises(AttributeError, match="can't set attribute"):
        ex.message = err_msg
    with pytest.raises(AttributeError, match="can't set attribute"):
        ex.error_code = err_code
    with pytest.raises(AttributeError, match="can't set attribute"):
        ex.error_type = err_type


# =======================================================================
#                       Class PipeJsonRpcReceive


def test_PipeJsonRpcReceive_1():
    """
    Create, start and stop `PipeJsonRpcReceive` object
    """
    conn1, conn2 = multiprocessing.Pipe()
    new_name = "Unusual Thread Name"

    assert count_threads_with_name(new_name) == 0, "No threads are expected to exist"

    pc = PipeJsonRpcReceive(conn=conn2, name=new_name)
    pc.start()
    assert count_threads_with_name(new_name) == 1, "One thread is expected to exist"

    pc.start()  # Expected to do nothing

    pc.stop()
    ttime.sleep(0.15)  # Wait until the thread stops (wait more than the 0.1s polling period)
    assert count_threads_with_name(new_name) == 0, "No threads are expected to exist"

    pc.start()  # Restart
    assert count_threads_with_name(new_name) == 1, "One thread is expected to exist"
    pc.stop()


@pytest.mark.parametrize(
    "method, params, result, notification",
    [
        ("method_handler1", [], 5, False),
        ("method_handler1", [], 5, True),
        ("method1", [], 5, False),
        ("method2", [5], 15, False),
        ("method2", {"value": 5}, 15, False),
        ("method2", {}, 12, False),
        ("method3", {"value": 5}, 20, False),
        ("method3", {}, 18, False),
        ("method4", {"value": 5}, 20, False),
        ("method4", {}, 19, False),
    ],
)
def test_PipeJsonRpcReceive_2(method, params, result, notification):
    """
    The case of single requests.
    """
    value_nonlocal = None

    def method_handler1():
        nonlocal value_nonlocal
        value_nonlocal = "function_was_called"
        return 5

    def method_handler2(value=2):
        nonlocal value_nonlocal
        value_nonlocal = "function_was_called"
        return value + 10

    def method_handler3(*, value=3):
        nonlocal value_nonlocal
        value_nonlocal = "function_was_called"
        return value + 15

    class SomeClass:
        def method_handler4(self, *, value=4):
            nonlocal value_nonlocal
            value_nonlocal = "function_was_called"
            return value + 15

    some_class = SomeClass()

    conn1, conn2 = multiprocessing.Pipe()
    pc = PipeJsonRpcReceive(conn=conn2)
    pc.add_method(method_handler1)  # No name is specified, default name is "method_handler1"
    pc.add_method(method_handler1, "method1")
    pc.add_method(method_handler2, "method2")
    pc.add_method(method_handler3, "method3")
    pc.add_method(some_class.method_handler4, "method4")
    pc.start()

    for n in range(3):
        value_nonlocal = None

        request = format_jsonrpc_msg(method, params, notification=notification)
        conn1.send(json.dumps(request))
        if conn1.poll(timeout=0.5):  # Set timeout large enough
            if not notification:
                response = conn1.recv()
                response = json.loads(response)
                assert response["id"] == request["id"], "Response ID does not match message ID."
                assert "result" in response, f"Key 'result' is not contained in response: {response}"
                assert response["result"] == result, f"Result does not match the expected: {response}"
                assert value_nonlocal == "function_was_called", "Non-local variable has incorrect value"
            else:
                assert False, "Notification was sent but response was received."
        else:
            # If request is a notification, there should be no response.
            if not notification:
                assert False, "Timeout occurred while waiting for response."
            else:
                assert value_nonlocal == "function_was_called", "Non-local variable has incorrect value"

    pc.stop()


def test_PipeJsonRpcReceive_3():
    """
    Test sending multiple requests
    """

    def method_handler3(*, value=3):
        return value + 15

    class SomeClass:
        def method_handler4(self, *, value=4):
            return value + 7

    some_class = SomeClass()

    conn1, conn2 = multiprocessing.Pipe()
    pc = PipeJsonRpcReceive(conn=conn2)
    pc.add_method(method_handler3, "method3")
    pc.add_method(some_class.method_handler4, "method4")
    pc.start()

    # Non-existing method ('Method not found' error)
    request = [
        format_jsonrpc_msg("method3", {"value": 7}),
        format_jsonrpc_msg("method4", {"value": 3}, notification=True),
        format_jsonrpc_msg("method4", {"value": 9}),
    ]
    conn1.send(json.dumps(request))

    if conn1.poll(timeout=0.5):  # Set timeout large enough
        response = conn1.recv()
        response = json.loads(response)
        assert len(response) == 2, "Unexpected number of response messages"
        assert response[0]["id"] == request[0]["id"], "Response ID does not match message ID."
        assert response[1]["id"] == request[2]["id"], "Response ID does not match message ID."
        assert response[0]["result"] == 22, "Response ID does not match message ID."
        assert response[1]["result"] == 16, "Response ID does not match message ID."
    else:
        assert False, "Timeout occurred while waiting for response."

    pc.stop()


def test_PipeJsonRpcReceive_4():
    """
    Test if all outdated unprocessed messages are deleted once the processing thread is started.
    """

    def method_handler3(*, value=3):
        return value + 15

    conn1, conn2 = multiprocessing.Pipe()
    pc = PipeJsonRpcReceive(conn=conn2)
    pc.add_method(method_handler3, "method3")

    # The thread is not started yet, but we still send a message through the pipe.
    #   This message is expected to be deleted once the thread starts.
    request1a = [
        format_jsonrpc_msg("method3", {"value": 5}),
    ]
    request1b = [
        format_jsonrpc_msg("method3", {"value": 6}),
    ]
    conn1.send(json.dumps(request1a))
    conn1.send(json.dumps(request1b))

    # Start the processing thread. The messages that were already sent are expected to be ignored.
    pc.start()

    # Send another message. This message is expected to be processed.
    request2 = [
        format_jsonrpc_msg("method3", {"value": 7}),
    ]
    conn1.send(json.dumps(request2))

    if conn1.poll(timeout=0.5):  # Set timeout large enough
        response = conn1.recv()
        response = json.loads(response)
        assert len(response) == 1, "Unexpected number of response messages"
        assert response[0]["id"] == request2[0]["id"], "Response ID does not match message ID."
        assert response[0]["result"] == 22, "Response ID does not match message ID."
    else:
        assert False, "Timeout occurred while waiting for response."

    pc.stop()


def test_PipeJsonRpcReceive_5_failing():
    """
    Those tests are a result of exploration of how `json-rpc` error processing works.
    """

    def method_handler3(*, value=3):
        return value + 15

    conn1, conn2 = multiprocessing.Pipe()
    pc = PipeJsonRpcReceive(conn=conn2)
    pc.add_method(method_handler3, "method3")
    pc.start()

    # ------- Incorrect argument (arg list instead of required kwargs) -------
    #   Returns 'Server Error' (-32000)
    request = format_jsonrpc_msg("method3", [5])
    conn1.send(json.dumps(request))

    if conn1.poll(timeout=0.5):  # Set timeout large enough
        response = conn1.recv()
        response = json.loads(response)
        assert response["error"]["code"] == -32000, f"Incorrect error reported: {response}"
        assert response["error"]["data"]["type"] == "TypeError", "Incorrect error type."
    else:
        assert False, "Timeout occurred while waiting for response."

    # ------- Incorrect argument (incorrect argument type) -------
    # 'json-prc' doesn't check parameter types. Instead the handler will crash if the argument
    #   type is not suitable.
    request = format_jsonrpc_msg("method3", {"value": "abc"})
    conn1.send(json.dumps(request))

    if conn1.poll(timeout=0.5):
        response = conn1.recv()
        response = json.loads(response)
        assert response["error"]["code"] == -32000, f"Incorrect error reported: {response}"
        assert response["error"]["data"]["type"] == "TypeError", "Incorrect error type."
    else:
        assert False, "Timeout occurred while waiting for response."

    # ------- Incorrect argument (extra argument) -------
    request = format_jsonrpc_msg("method3", {"value": 5, "unknown": 10})
    conn1.send(json.dumps(request))

    if conn1.poll(timeout=0.5):
        response = conn1.recv()
        response = json.loads(response)
        assert response["error"]["code"] == -32602, f"Incorrect error reported: {response}"
    else:
        assert False, "Timeout occurred while waiting for response."

    # ------- Non-existing method ('Method not found' error) -------
    request = format_jsonrpc_msg("method_handler3", {"value": 5})
    conn1.send(json.dumps(request))

    if conn1.poll(timeout=0.5):
        response = conn1.recv()
        response = json.loads(response)
        assert response["error"]["code"] == -32601, f"Incorrect error reported: {response}"
    else:
        assert False, "Timeout occurred while waiting for response."

    pc.stop()


def test_PipeJsonRpcReceive_6_failing():
    """
    Exception is raised inside the handler is causing 'Server Error' -32000.
    Returns error type (Exception type) and message. It is questionable whether
    built-in exception processing should be used to process exceptions within handlers.
    It may be better to catch and process all exceptions produced by the custom code and
    leave built-in processing for exceptions that happen while calling the function.

    EXCEPTIONS SHOULD BE PROCESSED INSIDE THE HANDLER!!!
    """

    def method_handler5():
        raise RuntimeError("Function crashed ...")

    conn1, conn2 = multiprocessing.Pipe()
    pc = PipeJsonRpcReceive(conn=conn2)
    pc.add_method(method_handler5, "method5")
    pc.start()

    request = format_jsonrpc_msg("method5")
    conn1.send(json.dumps(request))

    if conn1.poll(timeout=0.5):  # Set timeout large enough
        response = conn1.recv()
        response = json.loads(response)
        assert response["error"]["code"] == -32000, f"Incorrect error reported: {response}"
        assert response["error"]["data"]["type"] == "RuntimeError", "Incorrect error type."
        assert response["error"]["data"]["message"] == "Function crashed ...", "Incorrect message."
    else:
        assert False, "Timeout occurred while waiting for response."

    pc.stop()


def test_PipeJsonRpcReceive_7_failing():
    """
    This is an example of handler to timeout. 'json-rpc' package can not handle timeouts.
    Care must be taken to write handles that execute quickly. Timeouts must be handled
    explicitly within the handlers.

    ONLY INSTANTLY EXECUTED HANDLERS MAY BE USED!!!
    """

    def method_handler6():
        ttime.sleep(3)  # Longer than 'poll' timeout

    conn1, conn2 = multiprocessing.Pipe()
    pc = PipeJsonRpcReceive(conn=conn2)
    pc.add_method(method_handler6, "method6")
    pc.start()

    # Non-existing method ('Method not found' error)
    request = format_jsonrpc_msg("method6")

    conn1.send(json.dumps(request))

    if conn1.poll(timeout=0.5):  # Set timeout large enough
        assert False, "The test is expected to timeout."
    else:
        pass

    pc.stop()


# =======================================================================
#                       Class PipeJsonRpcSendAsync


def test_PipeJsonRpcSendAsync_1():
    """
    Create, start and stop `PipeJsonRpcReceive` object
    """
    conn1, conn2 = multiprocessing.Pipe()
    new_name = "Unusual Thread Name"

    async def object_start_stop():
        assert count_threads_with_name(new_name) == 0, "No threads are expected to exist"

        pc = PipeJsonRpcSendAsync(conn=conn1, name=new_name)
        pc.start()
        assert count_threads_with_name(new_name) == 1, "One thread is expected to exist"

        pc.start()  # Expected to do nothing

        pc.stop()
        ttime.sleep(0.15)  # Wait until the thread stops (0.1s polling period)
        assert count_threads_with_name(new_name) == 0, "No threads are expected to exist"

        pc.start()  # Restart
        assert count_threads_with_name(new_name) == 1, "One thread is expected to exist"
        pc.stop()

    asyncio.run(object_start_stop())


@pytest.mark.parametrize(
    "method, params, result, notification",
    [
        ("method_handler1", [], 5, False),
        ("method_handler1", [], 5, True),
        ("method1", [], 5, False),
        ("method2", [5], 15, False),
        ("method2", {"value": 5}, 15, False),
        ("method2", {}, 12, False),
        ("method3", {"value": 5}, 20, False),
        ("method3", {}, 18, False),
        ("method4", {"value": 5}, 20, False),
        ("method4", {}, 19, False),
    ],
)
def test_PipeJsonRpcSendAsync_2(method, params, result, notification):
    """
    Test of basic functionality. Here we don't test for timeout case (it raises an exception).
    """
    value_nonlocal = None

    def method_handler1():
        nonlocal value_nonlocal
        value_nonlocal = "function_was_called"
        return 5

    def method_handler2(value=2):
        nonlocal value_nonlocal
        value_nonlocal = "function_was_called"
        return value + 10

    def method_handler3(*, value=3):
        nonlocal value_nonlocal
        value_nonlocal = "function_was_called"
        return value + 15

    class SomeClass:
        def method_handler4(self, *, value=4):
            nonlocal value_nonlocal
            value_nonlocal = "function_was_called"
            return value + 15

    some_class = SomeClass()

    conn1, conn2 = multiprocessing.Pipe()
    pc = PipeJsonRpcReceive(conn=conn2, name="comm-server")
    pc.add_method(method_handler1)  # No name is specified, default name is "method_handler1"
    pc.add_method(method_handler1, "method1")
    pc.add_method(method_handler2, "method2")
    pc.add_method(method_handler3, "method3")
    pc.add_method(some_class.method_handler4, "method4")
    pc.start()

    async def send_messages():
        nonlocal value_nonlocal

        p_send = PipeJsonRpcSendAsync(conn=conn1, name="comm-client")
        p_send.start()

        for n in range(3):
            value_nonlocal = None

            response = await p_send.send_msg(method, params, notification=notification)
            if not notification:
                assert response == result, f"Result does not match the expected: {response}"
                assert value_nonlocal == "function_was_called", "Non-local variable has incorrect value"
            elif response is not None:
                assert False, "Response was received for notification."

        p_send.stop()

    asyncio.run(send_messages())
    pc.stop()


def test_PipeJsonRpcSendAsync_3():
    """
    Put multiple messages to the loop at once. The should be processed one by one.
    """
    n_calls = 0
    lock = threading.Lock()

    def method_handler1():
        nonlocal n_calls
        with lock:
            n_calls += 1
            n_return = n_calls
        ttime.sleep(0.1)
        return n_return

    conn1, conn2 = multiprocessing.Pipe()
    pc = PipeJsonRpcReceive(conn=conn2, name="comm-server")
    pc.add_method(method_handler1, "method1")
    pc.start()

    async def send_messages():
        p_send = PipeJsonRpcSendAsync(conn=conn1, name="comm-client")
        p_send.start()

        # Submit multiple messages at once. Messages should stay at the event loop
        #   and be processed one by one.
        futs = []
        for n in range(5):
            futs.append(asyncio.ensure_future(p_send.send_msg("method1")))

        for n, fut in enumerate(futs):
            await asyncio.wait_for(fut, timeout=5.0)  # Timeout is in case of failure
            result = fut.result()
            assert result == n + 1, "Incorrect returned value"

        p_send.stop()

    asyncio.run(send_messages())
    pc.stop()


def test_PipeJsonRpcSendAsync_4():
    """
    Message timeout.
    """

    def method_handler1():
        ttime.sleep(1)

    conn1, conn2 = multiprocessing.Pipe()
    pc = PipeJsonRpcReceive(conn=conn2, name="comm-server")
    pc.add_method(method_handler1, "method1")
    pc.start()

    async def send_messages():
        p_send = PipeJsonRpcSendAsync(conn=conn1, name="comm-client")
        p_send.start()

        # Submit multiple messages at once. Messages should stay at the event loop
        #   and be processed one by one.
        with pytest.raises(CommTimeoutError, match="Timeout while waiting for response to message"):
            await p_send.send_msg("method1", timeout=0.5)

        p_send.stop()

    asyncio.run(send_messages())
    pc.stop()


def test_PipeJsonRpcSendAsync_5():
    """
    Specia test case.
    Two messages: the first message times out, the second message is send before the response
    from the first message is received. Verify that the result returned in response to the
    second message is received. (We discard the result of the message that is timed out.)
    """

    def method_handler1():
        ttime.sleep(0.7)
        return 39

    def method_handler2():
        ttime.sleep(0.2)
        return 56

    conn1, conn2 = multiprocessing.Pipe()
    pc = PipeJsonRpcReceive(conn=conn2, name="comm-server")
    pc.add_method(method_handler1, "method1")
    pc.add_method(method_handler2, "method2")
    pc.start()

    async def send_messages():
        p_send = PipeJsonRpcSendAsync(conn=conn1, name="comm-client")
        p_send.start()

        # Submit multiple messages at once. Messages should stay at the event loop
        #   and be processed one by one.
        with pytest.raises(CommTimeoutError):
            await p_send.send_msg("method1", timeout=0.5)

        result = await p_send.send_msg("method2", timeout=0.5)
        assert result == 56, "Incorrect result received"

        p_send.stop()

    asyncio.run(send_messages())
    pc.stop()


def test_PipeJsonRpcSendAsync_6_fail():
    """
    Exception raised inside the method.
    """

    def method_handler1():
        raise RuntimeError("Function crashed ...")

    conn1, conn2 = multiprocessing.Pipe()
    pc = PipeJsonRpcReceive(conn=conn2, name="comm-server")
    pc.add_method(method_handler1, "method1")
    pc.start()

    async def send_messages():
        p_send = PipeJsonRpcSendAsync(conn=conn1, name="comm-client")
        p_send.start()

        # Submit multiple messages at once. Messages should stay at the event loop
        #   and be processed one by one.
        with pytest.raises(CommJsonRpcError, match="Function crashed ..."):
            await p_send.send_msg("method1", timeout=0.5)

        p_send.stop()

    asyncio.run(send_messages())
    pc.stop()


def test_PipeJsonRpcSendAsync_7_fail():
    """
    Method not found (other `json-rpc` errors will raise the same exception).
    """

    def method_handler1():
        pass

    conn1, conn2 = multiprocessing.Pipe()
    pc = PipeJsonRpcReceive(conn=conn2, name="comm-server")
    pc.add_method(method_handler1, "method1")
    pc.start()

    async def send_messages():
        p_send = PipeJsonRpcSendAsync(conn=conn1, name="comm-client")
        p_send.start()

        # Submit multiple messages at once. Messages should stay at the event loop
        #   and be processed one by one.
        with pytest.raises(CommJsonRpcError):
            await p_send.send_msg("nonexisting_method", timeout=0.5)

        p_send.stop()

    asyncio.run(send_messages())
    pc.stop()


# =======================================================================
#                               ZMQ keys


def test_generate_zmq_keys():
    """
    Functions ``generate_new_zmq_key_pair()``, ``generate_zmq_public_key()`` and ``validate_zmq_key()``.
    """
    key_public, key_private = generate_new_zmq_key_pair()
    assert isinstance(key_public, str)
    assert len(key_public) == 40
    assert isinstance(key_private, str)
    assert len(key_private) == 40

    validate_zmq_key(key_public)
    validate_zmq_key(key_private)

    key_public_gen = generate_zmq_public_key(key_private)
    assert key_public_gen == key_public


# fmt: off
@pytest.mark.parametrize("key", [
    None,
    10,
    "",
    "abc",
    "wt8[6a8eoXFRVL<l2JBbOzs(hcI%kRBIr0Do/eL'",  # 40 characters, but contains invalid character "'"
])
# fmt: on
def test_validate_zmq_key(key):
    """
    Function ``validate_zmq_key()``: cases of failing validation.
    """
    with pytest.raises(ValueError, match="the key must be a 40 byte z85 encoded string"):
        validate_zmq_key(key)


# =======================================================================
#                       Class ZMQCommSendThreads


def _gen_server_keys(*, encryption_enabled):
    """Generate server keys and a set of kwargs."""
    if encryption_enabled:
        public_key, private_key = generate_new_zmq_key_pair()
        server_kwargs = {"private_key": private_key}
    else:
        public_key, private_key = None, None
        server_kwargs = {}
    return public_key, private_key, server_kwargs


def _zmq_server_1msg(*, private_key=None):
    """
    ZMQ server that provides single response.
    ``private_key`` - server private key (for tests with enabled encryption)
    """
    ctx = zmq.Context()
    zmq_socket = ctx.socket(zmq.REP)

    if private_key is not None:
        zmq_socket.set(zmq.CURVE_SERVER, 1)
        zmq_socket.set(zmq.CURVE_SECRETKEY, private_key.encode("utf-8"))

    zmq_socket.bind("tcp://*:60615")
    msg_in = zmq_socket.recv_json()
    msg_out = {"success": True, "some_data": 10, "msg_in": msg_in}
    zmq_socket.send_json(msg_out)
    zmq_socket.close(linger=10)


# fmt: off
@pytest.mark.parametrize("encryption_enabled", [False, True])
@pytest.mark.parametrize("is_blocking", [True, False])
# fmt: on
def test_ZMQCommSendThreads_1(is_blocking, encryption_enabled):
    """
    Basic test of ZMQCommSendThreads class: single communication with the
    server both in blocking and non-blocking mode.
    """
    public_key, _, server_kwargs = _gen_server_keys(encryption_enabled=encryption_enabled)

    thread = threading.Thread(target=_zmq_server_1msg, kwargs=server_kwargs)
    thread.start()

    zmq_comm = ZMQCommSendThreads(server_public_key=public_key if encryption_enabled else None)
    method, params = "testing", {"p1": 10, "p2": "abc"}

    msg_recv, msg_recv_err = {}, ""

    if is_blocking:
        msg_recv = zmq_comm.send_message(method=method, params=params)
    else:
        done = False

        def cb(msg, msg_err):
            nonlocal msg_recv, msg_recv_err, done
            msg_recv = msg
            msg_recv_err = msg_err
            done = True

        zmq_comm.send_message(method=method, params=params, cb=cb)
        # Implement primitive polling of 'done' flag
        while not done:
            ttime.sleep(0.1)

    assert msg_recv["success"] is True, str(msg_recv)
    assert msg_recv["some_data"] == 10, str(msg_recv)
    assert msg_recv["msg_in"] == {"method": method, "params": params}, str(msg_recv)
    assert msg_recv_err == ""

    thread.join()


def _zmq_server_2msg(*, private_key=None):
    """
    ZMQ server: provides ability to communicate twice.
    ``private_key`` - server private key (for tests with enabled encryption)
    """
    ctx = zmq.Context()
    zmq_socket = ctx.socket(zmq.REP)

    if private_key is not None:
        zmq_socket.set(zmq.CURVE_SERVER, 1)
        zmq_socket.set(zmq.CURVE_SECRETKEY, private_key.encode("utf-8"))

    zmq_socket.bind("tcp://*:60615")
    msg_in = zmq_socket.recv_json()
    msg_out = {"success": True, "some_data": 10, "msg_in": msg_in}
    zmq_socket.send_json(msg_out)
    msg_in = zmq_socket.recv_json()
    msg_out = {"success": True, "some_data": 20, "msg_in": msg_in}
    zmq_socket.send_json(msg_out)
    zmq_socket.close(linger=10)


# fmt: off
@pytest.mark.parametrize("encryption_enabled", [False, True])
@pytest.mark.parametrize("is_blocking", [True, False])
# fmt: on
def test_ZMQCommSendThreads_2(is_blocking, encryption_enabled):
    """
    Basic test of ZMQCommSendThreads class: two consecutive communications with the
    server both in blocking and non-blocking mode.
    """
    public_key, _, server_kwargs = _gen_server_keys(encryption_enabled=encryption_enabled)

    thread = threading.Thread(target=_zmq_server_2msg, kwargs=server_kwargs)
    thread.start()

    zmq_comm = ZMQCommSendThreads(server_public_key=public_key if encryption_enabled else None)
    method, params = "testing", {"p1": 10, "p2": "abc"}

    msg_recv, msg_recv_err = {}, ""

    for val in (10, 20):
        if is_blocking:
            msg_recv = zmq_comm.send_message(method=method, params=params)
        else:
            done = False

            def cb(msg, msg_err):
                nonlocal msg_recv, msg_recv_err, done
                msg_recv = msg
                msg_recv_err = msg_err
                done = True

            zmq_comm.send_message(method=method, params=params, cb=cb)
            # Implement primitive polling of 'done' flag
            while not done:
                ttime.sleep(0.1)

        assert msg_recv["success"] is True, str(msg_recv)
        assert msg_recv["some_data"] == val, str(msg_recv)
        assert msg_recv["msg_in"] == {"method": method, "params": params}, str(msg_recv)
        assert msg_recv_err == ""

    thread.join()


def _zmq_server_2msg_delay1(*, private_key=None):
    """
    ZMQ server: provides ability to communicate twice, short delay before sending the 1st response
    ``private_key`` - server private key (for tests with enabled encryption)
    """
    ctx = zmq.Context()
    zmq_socket = ctx.socket(zmq.REP)

    if private_key is not None:
        zmq_socket.set(zmq.CURVE_SERVER, 1)
        zmq_socket.set(zmq.CURVE_SECRETKEY, private_key.encode("utf-8"))

    zmq_socket.bind("tcp://*:60615")
    msg_in = zmq_socket.recv_json()
    ttime.sleep(0.1)  # Delay before the 1st response
    msg_out = {"success": True, "some_data": 10, "msg_in": msg_in}
    zmq_socket.send_json(msg_out)
    msg_in = zmq_socket.recv_json()
    msg_out = {"success": True, "some_data": 20, "msg_in": msg_in}
    zmq_socket.send_json(msg_out)
    zmq_socket.close(linger=10)


# fmt: off
@pytest.mark.parametrize("encryption_enabled", [False, True])
@pytest.mark.parametrize("is_blocking", [True, False])
# fmt: on
def test_ZMQCommSendThreads_3(is_blocking, encryption_enabled):
    """
    Testing protection of '_zmq_communicate` with lock. In this test the function
    ``send_message` is called twice so that the second call is submitted before
    the response to the first message is received. The server waits for 0.1 seconds
    before responding to the 1st message to emulate delay in processing. Since
    ``_zmq_communicate`` is protected by a lock, the second request will not
    be sent until the first message is processed.
    """
    public_key, _, server_kwargs = _gen_server_keys(encryption_enabled=encryption_enabled)

    thread = threading.Thread(target=_zmq_server_2msg_delay1, kwargs=server_kwargs)
    thread.start()

    zmq_comm = ZMQCommSendThreads(server_public_key=public_key if encryption_enabled else None)
    method, params = "testing", {"p1": 10, "p2": "abc"}

    msg_recv, msg_recv_err = [], []
    n_done = 0

    def cb(msg, msg_err):
        nonlocal msg_recv, msg_recv_err, n_done
        msg_recv.append(msg)
        msg_recv_err.append(msg_err)
        n_done += 1

    vals = (10, 20)

    if is_blocking:
        # In case of blocking call the lock can only be tested using threads
        def thread_request():
            nonlocal msg_recv, msg_recv_err, n_done
            _ = zmq_comm.send_message(method=method, params=params)
            msg_recv.append(_)
            msg_recv_err.append("")
            n_done += 1

        th_request = threading.Thread(target=thread_request)
        th_request.start()

        # Call the same function in the main thread (send 2nd message to the server)
        thread_request()

        th_request.join()

    else:
        for n in range(len(vals)):
            zmq_comm.send_message(method=method, params=params, cb=cb)

    while n_done < 2:
        ttime.sleep(0.1)

    for n, val in enumerate(vals):
        assert msg_recv[n]["success"] is True, str(msg_recv)
        assert msg_recv[n]["some_data"] == val, str(msg_recv)
        assert msg_recv[n]["msg_in"] == {"method": method, "params": params}, str(msg_recv)
        assert msg_recv_err[n] == ""

    thread.join()


def _zmq_server_delay2(*, private_key=None):
    """
    ZMQ server: provides ability to communicate twice, long delay before sending the 1st response
    ``private_key`` - server private key (for tests with enabled encryption)
    """
    ctx = zmq.Context()
    zmq_socket = ctx.socket(zmq.REP)

    if private_key is not None:
        zmq_socket.set(zmq.CURVE_SERVER, 1)
        zmq_socket.set(zmq.CURVE_SECRETKEY, private_key.encode("utf-8"))

    zmq_socket.bind("tcp://*:60615")
    msg_in = zmq_socket.recv_json()
    ttime.sleep(3)  # Generate timeout at the client
    msg_out = {"success": True, "some_data": 10, "msg_in": msg_in}
    zmq_socket.send_json(msg_out)
    msg_in = zmq_socket.recv_json()
    msg_out = {"success": True, "some_data": 20, "msg_in": msg_in}
    zmq_socket.send_json(msg_out)
    zmq_socket.close(linger=10)


# fmt: off
@pytest.mark.parametrize("encryption_enabled", [False, True])
@pytest.mark.parametrize(
    "is_blocking, raise_exception",
    [(True, False),  # Repeated test are intentional
     (True, False),
     (True, None),
     (True, True),
     (False, False),
     (False, False)]
)
@pytest.mark.parametrize("delay_between_reads", [2, 0.1])
# fmt: on
def test_ZMQCommSendThreads_4(is_blocking, raise_exception, delay_between_reads, encryption_enabled):
    """
    ZMQCommSendThreads: Timeout at the server.
    """
    public_key, _, server_kwargs = _gen_server_keys(encryption_enabled=encryption_enabled)

    thread = threading.Thread(target=_zmq_server_delay2, kwargs=server_kwargs)
    thread.start()

    zmq_comm = ZMQCommSendThreads(server_public_key=public_key if encryption_enabled else None)
    method, params = "testing", {"p1": 10, "p2": "abc"}

    msg_recv, msg_recv_err = {}, ""

    for val in (10, 20):
        if is_blocking:
            if (raise_exception in (True, None)) and (val == 10):
                # The case when timeout is expected for blocking operation
                with pytest.raises(CommTimeoutError, match="timeout occurred"):
                    zmq_comm.send_message(method=method, params=params, raise_exceptions=raise_exception)
            else:
                msg_recv = zmq_comm.send_message(method=method, params=params, raise_exceptions=raise_exception)
        else:
            done = False

            def cb(msg, msg_err):
                nonlocal msg_recv, msg_recv_err, done
                msg_recv = msg
                msg_recv_err = msg_err
                done = True

            zmq_comm.send_message(method=method, params=params, cb=cb, raise_exceptions=raise_exception)

            # Implement primitive polling of 'done' flag
            while not done:
                ttime.sleep(0.1)

        if val == 10:
            if is_blocking:
                if raise_exception not in (True, None):
                    assert msg_recv["success"] is False, str(msg_recv)
                    assert "timeout occurred" in msg_recv["msg"], str(msg_recv)
            else:
                assert msg_recv == {}
                assert "timeout occurred" in msg_recv_err

            # Delay between consecutive reads. Test cases when read is initiated before and
            #   after the server restored operation and sent the response.
            ttime.sleep(delay_between_reads)

        else:
            assert msg_recv["success"] is True, str(msg_recv)
            assert msg_recv["some_data"] == val, str(msg_recv)
            assert msg_recv["msg_in"] == {"method": method, "params": params}, str(msg_recv)
            assert msg_recv_err == ""

    thread.join()


def test_ZMQCommSendThreads_5_fail():
    """
    Invalid public key
    """
    with pytest.raises(ValueError):
        ZMQCommSendThreads(server_public_key="abc")


# =======================================================================
#                       Class ZMQCommSendAsync

# fmt: off
@pytest.mark.parametrize("encryption_enabled", [False, True])
# fmt: on
def test_ZMQCommSendAsync_1(encryption_enabled):
    """
    Basic test of ZMQCommSendAsync class: send one message to the server and receive response.
    """
    public_key, _, server_kwargs = _gen_server_keys(encryption_enabled=encryption_enabled)

    thread = threading.Thread(target=_zmq_server_1msg, kwargs=server_kwargs)
    thread.start()

    async def testing():
        zmq_comm = ZMQCommSendAsync(server_public_key=public_key if encryption_enabled else None)
        method, params = "testing", {"p1": 10, "p2": "abc"}

        msg_recv = await zmq_comm.send_message(method=method, params=params)

        assert msg_recv["success"] is True, str(msg_recv)
        assert msg_recv["some_data"] == 10, str(msg_recv)
        assert msg_recv["msg_in"] == {"method": method, "params": params}, str(msg_recv)

    asyncio.run(testing())

    thread.join()


# fmt: off
@pytest.mark.parametrize("encryption_enabled", [False, True])
# fmt: on
def test_ZMQCommSendAsync_2(encryption_enabled):
    """
    Basic test of ZMQCommSendAsync class: two consecutive communications with the
    server both in blocking and non-blocking mode.
    """
    public_key, _, server_kwargs = _gen_server_keys(encryption_enabled=encryption_enabled)

    thread = threading.Thread(target=_zmq_server_2msg, kwargs=server_kwargs)
    thread.start()

    async def testing():
        zmq_comm = ZMQCommSendAsync(server_public_key=public_key if encryption_enabled else None)
        method, params = "testing", {"p1": 10, "p2": "abc"}

        for val in (10, 20):
            msg_recv = await zmq_comm.send_message(method=method, params=params)

            assert msg_recv["success"] is True, str(msg_recv)
            assert msg_recv["some_data"] == val, str(msg_recv)
            assert msg_recv["msg_in"] == {"method": method, "params": params}, str(msg_recv)

    asyncio.run(testing())

    thread.join()


# fmt: off
@pytest.mark.parametrize("encryption_enabled", [False, True])
# fmt: on
def test_ZMQCommSendAsync_3(encryption_enabled):
    """
    Testing protection of '_zmq_communicate` with lock. In this test the function
    ``send_message` is called twice so that the second call is submitted before
    the response to the first message is received. The server waits for 0.1 seconds
    before responding to the 1st message to emulate delay in processing. Since
    ``_zmq_communicate`` is protected by a lock, the second request will not
    be sent until the first message is processed.
    """
    public_key, _, server_kwargs = _gen_server_keys(encryption_enabled=encryption_enabled)

    thread = threading.Thread(target=_zmq_server_2msg_delay1, kwargs=server_kwargs)
    thread.start()

    async def testing():

        zmq_comm = ZMQCommSendAsync(server_public_key=public_key if encryption_enabled else None)
        method, params = "testing", {"p1": 10, "p2": "abc"}

        vals = (10, 20)

        tasks = []
        for n in range(len(vals)):
            tsk = asyncio.ensure_future(zmq_comm.send_message(method=method, params=params))
            tasks.append(tsk)

        for tsk in tasks:
            await tsk

        for n, val in enumerate(vals):
            res = tasks[n].result()
            assert res["success"] is True, str(res)
            assert res["some_data"] == val, str(res)
            assert res["msg_in"] == {"method": method, "params": params}, str(res)

    asyncio.run(testing())

    thread.join()


# fmt: off
@pytest.mark.parametrize("encryption_enabled", [False, True])
@pytest.mark.parametrize(
    "raise_exception",
    [False,  # Repeated tests are intentional
     False,
     None,
     True,
     False,
     False]
)
@pytest.mark.parametrize("delay_between_reads", [2, 0.1])
# fmt: on
def test_ZMQCommSendAsync_4(raise_exception, delay_between_reads, encryption_enabled):
    """
    ZMQCommSendAsync: Timeout at the server.
    """
    public_key, _, server_kwargs = _gen_server_keys(encryption_enabled=encryption_enabled)

    thread = threading.Thread(target=_zmq_server_delay2, kwargs=server_kwargs)
    thread.start()

    async def testing():

        zmq_comm = ZMQCommSendAsync(server_public_key=public_key if encryption_enabled else None)
        method, params = "testing", {"p1": 10, "p2": "abc"}

        msg_recv, msg_recv_err = {}, ""

        for val in (10, 20):
            if (raise_exception in (True, None)) and (val == 10):
                # The case when timeout is expected for blocking operation
                with pytest.raises(CommTimeoutError, match="timeout occurred"):
                    await zmq_comm.send_message(method=method, params=params, raise_exceptions=raise_exception)
            else:
                msg_recv = await zmq_comm.send_message(
                    method=method, params=params, raise_exceptions=raise_exception
                )

            if val == 10:
                if raise_exception not in (True, None):
                    assert msg_recv["success"] is False, str(msg_recv)
                    assert "timeout occurred" in msg_recv["msg"], str(msg_recv)

                # Delay between consecutive reads. Test cases when read is initiated before and
                #   after the server restored operation and sent the response.
                await asyncio.sleep(delay_between_reads)

            else:
                assert msg_recv["success"] is True, str(msg_recv)
                assert msg_recv["some_data"] == val, str(msg_recv)
                assert msg_recv["msg_in"] == {"method": method, "params": params}, str(msg_recv)
                assert msg_recv_err == ""

    asyncio.run(testing())

    thread.join()


def test_ZMQCommSendAsync_5_fail():
    """
    Invalid public key
    """

    async def testing():
        with pytest.raises(ValueError):
            ZMQCommSendAsync(server_public_key="abc")

    asyncio.run(testing())



import threading
import pprint
import json
import asyncio
import uuid
import zmq
import zmq.asyncio
from jsonrpc import JSONRPCResponseManager
from jsonrpc.dispatcher import Dispatcher

import logging

logger = logging.getLogger(__name__)

_fixed_public_key = "wt8[6a8eoXFRVL<l2JBbOzs(hcI%kRBIr0Do/eLC"
_fixed_private_key = "=@e7WwVuz{*eGcnv{AL@x2hmX!z^)wP3vKsQ{S7s"


class CommTimeoutError(TimeoutError):
    """
    Raised when communication error occurs
    """

    pass


class CommJsonRpcError(RuntimeError):
    """
    Raised when returned json-rpc message contains error

    Parameters
    ----------
    message: str
        Error message
    error_code: int
        Error code (returned by `json-rpc`)
    error_type: str
        Error type (returned by `json-rpc` or set to `'CommJsonRpcError'`)
    """

    def __init__(self, message, error_code, error_type):
        super().__init__(message)
        # TODO: change 'code' and 'type' to read-only properties
        self.__error_code = error_code
        self.__error_type = error_type

    @property
    def error_code(self):
        return self.__error_code

    @property
    def error_type(self):
        return self.__error_type

    @property
    def message(self):
        return super().__str__()

    def __str__(self):
        msg = super().__str__() + f"\nError code: {self.error_code}. Error type: {self.error_type}"
        return msg

    def __repr__(self):
        return f"CommJsonRpcError('{self.message}', {self.error_code}, '{self.error_type}')"


def format_jsonrpc_msg(method, params=None, *, notification=False):
    """
    Returns dictionary that contains JSON RPC message.

    Parameters
    ----------
    method: str
        Method name
    params: dict or list, optional
        List of args or dictionary of kwargs.
    notification: boolean
        If the message is notification, no response will be expected.
    """
    msg = {"method": method, "jsonrpc": "2.0"}
    if params is not None:
        msg["params"] = params
    if not notification:
        msg["id"] = str(uuid.uuid4())
    return msg


class PipeJsonRpcReceive:
    """
    The class contains functions for receiving and processing JSON RPC messages received on
    communication pipe.

    Parameters
    ----------
    conn: multiprocessing.Connection
        Reference to bidirectional end of a pipe (multiprocessing.Pipe)
    name: str
        Name of the receiving thread (it is better to assign meaningful unique names to threads.

    Examples
    --------

    .. code-block:: python

        conn1, conn2 = multiprocessing.Pipe()
        pc = PipeJsonRPC(conn=conn1, name="RE QServer Receive")

        def func():
            print("Testing")

        pc.add_handler(func, "some_method")
        pc.start()   # Wait and process commands
        # The function 'func' is called when the message with method=="some_method" is received
        pc.stop()  # Stop before exit to stop the thread.
    """

    def __init__(self, conn, *, name="RE QServer Comm"):
        self._conn = conn
        self._dispatcher = Dispatcher()  # json-rpc dispatcher
        self._thread_running = False  # Set True to exit the thread

        self._thread_name = name

        self._conn_polling_timeout = 0.1  # in sec.

    def start(self):
        """
        Start processing of the pipe messages
        """
        self._start_conn_thread()

    def stop(self):
        """
        Stop processing of the pipe messages (and exit the tread)
        """
        self._thread_running = False

    def __del__(self):
        self.stop()

    def add_method(self, handler, name=None):
        """
        Add method to json-rpc dispatcher.

        Parameters
        ----------
        handler: callable
            Reference to a handler
        name: str, optional
            Name to register (default is the handler name)
        """
        # Add method to json-rpc dispatcher
        self._dispatcher.add_method(handler, name)

    def _start_conn_thread(self):
        if not self._thread_running:

            # Clear the pipe from outdated unprocessed messages.
            while self._conn.poll():
                self._conn.recv()

            self._thread_running = True
            self._thread_conn = threading.Thread(
                target=self._receive_conn_thread, name=self._thread_name, daemon=True
            )
            self._thread_conn.start()

    def _receive_conn_thread(self):
        while True:
            if self._conn.poll(self._conn_polling_timeout):
                try:
                    msg = self._conn.recv()
                    # Messages should be handled in the event loop
                    self._conn_received(msg)
                except Exception as ex:
                    logger.exception(
                        "Exception occurred while waiting for RE Manager-> Watchdog message: %s", str(ex)
                    )
                    break
            if not self._thread_running:  # Exit thread
                break

    def _conn_received(self, msg):

        # if logger.level < 11:  # Print output only if logging level is DEBUG (10) or less
        #     msg_json = json.loads(msg)
        #     We don't want to print 'heartbeat' messages
        #     if not isinstance(msg_json, dict) or (msg_json["method"] != "heartbeat"):
        #         logger.debug("Command received RE Manager->Watchdog: %s", pprint.pformat(msg_json))

        response = JSONRPCResponseManager.handle(msg, self._dispatcher)
        if response:
            response = response.json
            self._conn.send(response)


class PipeJsonRpcSendAsync:
    """
    The class contains functions for supporting asyncio-based client for JSON RPC comminucation
    using interprocess communication pipe. The class object must be created on the loop (from one of
    `async` functions). This implementation allows calls only to one method at a time. Multiple
    `send_msg` requests may be put on the loop, but the next message is never sent before
    the response to the previous message is received or timeout occurred.

    Parameters
    ----------
    conn: multiprocessing.Connection
        Reference to bidirectional end of a pipe (multiprocessing.Pipe)
    timeout: float
        Default value of timeout: maximum time to wait for response after a message is sent
    name: str
        Name of the receiving thread (it is better to assign meaningful unique names to threads.

    Examples
    --------

    .. code-block:: python

        conn1, conn2 = multiprocessing.Pipe()

        async def send_messages():
            # Must be instantiated and used within the loop
            p_send = PipeJsonRpcSendAsync(conn=conn1, name="comm-client")
            p_send.start()

            method = "method_name"
            params = {"value": 10}   #   or list of args [10, 25]
            response = await p_send.send_msg(method, params, notification=notification)

            p_send.stop()

        asyncio.run(send_messages())
        pc.stop()


        pc = PipeJsonRpcSendAsync(conn=conn1, name="RE QServer Receive")

        def func():
            print("Testing")

        pc.add_handler(func, "some_method")
        pc.start()   # Wait and process commands
        # The function 'func' is called when the message with method=="some_method" is received
        pc.stop()  # Stop before exit to stop the thread.

    """

    def __init__(self, conn, *, timeout=0.5, name="RE QServer Comm"):
        self._conn = conn
        self._loop = asyncio.get_running_loop()

        self._thread_name = name

        self._fut_comm = None  # Future for waiting for messages from watchdog
        # Lock that prevents sending of the next message before response
        #   to the previous message is received.
        self._lock_comm = asyncio.Lock()
        self._timeout_comm = timeout  # Timeout (time to wait for response to a message)

        # Polling timeout for the pipe. The data will be read from the pipe instantly once it is available.
        #   The timeout determines how long it would take to stop the thread when needed.
        self._conn_polling_timeout = 0.1

        self._thread_running = False  # True - thread is running

        # Expected ID of the received message. The ID must be the same as the ID of the sent message.
        #   Ignore all message that don't have matching ID or no ID.
        self._expected_msg_id = None

    def start(self):
        """
        Start processing of the pipe messages
        """
        self._start_conn_thread()

    def stop(self):
        """
        Stop processing of the pipe messages (and exit the tread)
        """
        self._thread_running = False

    def __del__(self):
        self.stop()

    def _start_conn_thread(self):
        # Start 'receive' thread
        if not self._thread_running:
            self._thread_running = True
            self._pipe_receive_thread = threading.Thread(
                target=self._pipe_receive, name=self._thread_name, daemon=True
            )
            self._pipe_receive_thread.start()

    async def send_msg(self, method, params=None, *, notification=False, timeout=None):
        """
        Send JSON RPC message to server and return the result of the function (method)
        or raise exception in case of an error. Returns None if the message is notification.

        Parameters
        ----------
        method: str
            name of JSON RPC method
        params: list or dict
            args or kwargs of the remote method
        notification: boolean
            True - message is notification. The function returns immediately without
            waiting of the response, which is never generated for notification.
        timeout: float
            Timeout in seconds. If no response is received at expiration of timeout,
            `CommTimeoutError` is raised. If the response will be received later, it
            will be ignored.

        Raises
        ------
        CommTimeoutError
            Timeout occurred. Response is not received in time
        CommJsonRpcError
            Error occurred while processing the message. This could indicate an error
            in `json-rpc` package (e.g. method not found) or exception raised by
            the method itself. It is recommended that the methods catch and process
            their exceptions (may be except parameter validation) and leave
            `CommJsonRpcError` for reporting `json-rpc` errors. In well tested
            program this exception should never be raised.
        RuntimeError
            Unrecognized message received (message doesn't contain `result` or `error`
            keys. This should never happen in well tested program.

        The function will raise `CommTimeoutError` in case of communication timeout
        """
        # The lock protects from sending the next message
        #   before response to the previous message is received.

        if timeout is None:
            timeout = self._timeout_comm

        async with self._lock_comm:
            msg = format_jsonrpc_msg(method, params, notification=notification)
            try:
                msg_json = json.dumps(msg)
                self._conn.send(msg_json)

                # No response is expected if this is a notification
                if not notification:
                    self._expected_msg_id = msg["id"]
                    self._fut_comm = self._loop.create_future()
                    # Waiting for the future may raise 'asyncio.TimeoutError'
                    await asyncio.wait_for(self._fut_comm, timeout=timeout)
                    response = self._fut_comm.result()

                    if "result" in response:
                        return response["result"]
                    elif "error" in response:
                        # TODO: verify that this is all information that should be saved
                        err_code = response["error"]["code"]
                        if "data" in response["error"]:
                            # Server Error (issue with execution of the method)
                            err_type = response["error"]["data"]["type"]
                            # Message: "Server error: <message text>"
                            err_msg = f'{response["error"]["message"]}: {response["error"]["data"]["message"]}'
                        else:
                            # Other json-rpc errors
                            err_type = "CommJsonRpcError"
                            err_msg = response["error"]["message"]
                        raise CommJsonRpcError(err_msg, error_code=err_code, error_type=err_type)
                    else:
                        err_msg = (
                            f"Message {pprint.pformat(msg)}\n"
                            f"resulted in response with unknown format: {pprint.pformat(response)}"
                        )
                        raise RuntimeError(err_msg)
                else:
                    response = None
                return response

            except asyncio.TimeoutError:
                raise CommTimeoutError(f"Timeout while waiting for response to message: \n{pprint.pformat(msg)}")
            finally:
                self._fut_comm = None
                self._expected_msg_id = None

    async def _response_received(self, response):
        """
        Set the future with the results. Ignore all messages with unexpected or missing IDs.
        Also ignore all unexpected messages.
        """
        if self._expected_msg_id is not None:
            if "id" in response:
                if response["id"] != self._expected_msg_id:
                    # Incorrect ID: ignore the message.
                    logger.error(
                        "Received response with incorrect message ID: %s. Expected %s.\nMessage: %s",
                        response["id"],
                        self._expected_msg_id,
                        pprint.pformat(response),
                    )
                else:
                    # Accept the message. Otherwise wait for timeout
                    self._fut_comm.set_result(response)
            else:
                # Missing ID: ignore the message
                logger.error("Received response with missing message ID: %s", pprint.pformat(response))
        else:
            logger.error(
                "Unsolicited message received: %s. Message is ignored",
                pprint.pformat(response),
            )

    def _conn_received(self, response):
        asyncio.create_task(self._response_received(response))

    def _pipe_receive(self):
        while True:
            if self._conn.poll(self._conn_polling_timeout):
                try:
                    msg_json = self._conn.recv()
                    msg = json.loads(msg_json)
                    # logger.debug("Message Watchdog->Manager received: '%s'", pprint.pformat(msg))
                    # Messages should be handled in the event loop
                    self._loop.call_soon_threadsafe(self._conn_received, msg)
                except Exception as ex:
                    logger.exception("Exception occurred while waiting for packet: %s", str(ex))
                    break
            if not self._thread_running:  # Exit thread
                break


# =========================================================================================
#                      Generation and validation of ZMQ key pairs


def generate_new_zmq_key_pair():
    """
    Generates new public-private key pair for ZMQ encryption.

    Returns
    -------
    str, str
        Public (first) and private (second) keys.
    """
    key_public, key_private = zmq.curve_keypair()
    return key_public.decode("utf-8"), key_private.decode("utf-8")


def generate_zmq_public_key(key_private):
    """
    Generates public key based on private key for ZMQ encryption.

    Returns
    -------
    str
        Public key.
    """
    key_public = zmq.curve_public(key_private.encode("utf-8"))
    return key_public.decode("utf-8")


def validate_zmq_key(key):
    """
    Validates format of a public or private key by feeding it to the function that generates
    public key from a private key. The function will raise an exception if the key is improperly
    formatted. The key is expected to be encoded using z85 (https://rfc.zeromq.org/spec/32/)
    and represented as a 40 character string.

    Parameters
    ----------
    key: str
        public or private key.

    Raises
    ------
    ValueError
        raised in case the key is not valid.
    """
    try:
        generate_zmq_public_key(key)
    except Exception:
        raise ValueError(f"Invalid key '{key}': the key must be a 40 byte z85 encoded string")


# =========================================================================================
#                                    ZMQ communication


class ZMQCommSendThreads:
    """
    Thread-based API for communication with RE Manager via ZMQ.

    Parameters
    ----------
    zmq_server_address : str or None
        Address of ZMQ server. If None, then the default address is ``tcp://localhost:60615``
        is used.
    timeout_recv : int
        Timeout (in ms) for ZMQ receive operations.
    timeout_send : int
        Timeout (in ms) for ZMQ send operations.
    raise_exceptions : bool
        Tells if exceptions should be raised in case of communication errors (mostly timeouts)
        when ``send_message()`` is called in blocking mode. This setting can be overridden by
        specifying ``raise_exceptions`` parameter of ``send_message()`` function. The exception
        ``CommTimeoutError`` is raised if the parameter is ``True``, otherwise error message
        is returned by ``send_message()``.
    server_public_key : str or None
        Server public key (z85-encoded 40 character string). The valid public key from the server
        public/private key pair must be passed if encryption is enabled at the 0MQ server side.
        Communication requests will time out if the key is invalid. Exception will be raised if
        the key is improperly formatted. Encryption will be disabled if ``None`` is passed.

    Examples
    --------

    .. code-block:: python

        zmq_comm = ZMQCommSendThreads()

        # Blocking call
        try:
            msg = zmq_comm.send_message(method="some_method", params={"some_value": n})
            # Code that uses msg
        except CommTimeoutError as ex:
            logger.exception("Exception occurred: %s", str(ex)

        # Non-blocking call (trivial example)
        msg_received, msg_err_received = [], []

        def cb(msg, msg_err):
            # msg - dict of parameters ({} if communication failed,
            # msg_err - string ("" if communication is successful)

            # In QT application, 'cb' would typically send a signal. There should be
            #   very limited amount of processing done in the callback, since it would
            #   block ZMQ communication.

            msg_received.append(msg)
            msg_err_received.append(msg_err)

        zmq_comm.send_message(method="some_method", params={"some_value": n}, cb=cb)

        # Wait for the message.
        while not msg_received:
            time.sleep(0.01)

        # Code to process received 'msg'
    """

    def __init__(
        self,
        *,
        zmq_server_address=None,
        timeout_recv=2000,
        timeout_send=500,
        raise_exceptions=True,
        server_public_key=None,
    ):

        if server_public_key is not None:
            validate_zmq_key(server_public_key)

        zmq_server_address = zmq_server_address or "tcp://localhost:60615"
        self._server_public_key = server_public_key

        self._timeout_receive = timeout_recv  # Timeout for 'recv' operation (ms)
        self._timeout_send = timeout_send  # # Timeout for 'send' operation (ms)
        self._raise_exceptions = raise_exceptions

        # ZeroMQ communication
        self._ctx = zmq.Context()
        self._zmq_socket = None
        self._zmq_server_address = zmq_server_address

        self._zmq_socket_open()
        self._lock_zmq = threading.Lock()

        self._blocking_call = False
        self._event_wait_for_msg = threading.Event()
        self._event_msg_received = threading.Event()

        self._zmq_error_occurred = False  # If True, then restart the socket
        self._cb = None  # Reference to callback function (used by 'send_message()')
        self._received_data = {}  # Keeps received data (used by 'send_message()')
        self._polling_timeout = 100  # Timeout used for polling the socket
        self._thread_name = "QServer ZMQ API"
        self._zmq_receive_thread = None  # Ref. to 'threading.Thread' object
        self._thread_running = False  # True - thread is running. Set False to exit the thread.

        # Start the thread. The thread will poll the socket for incoming data.
        self._start_receive_thread()

    def __del__(self):
        self.close()

    def _receive_thread(self):
        """
        Function that is executed in the 'receive' thread.
        """
        while True:
            if self._event_wait_for_msg.wait(timeout=self._polling_timeout / 1000):
                msg, msg_err = {}, ""
                try:
                    if self._zmq_socket.poll(timeout=self._timeout_receive):
                        msg = self._zmq_socket.recv_json()
                    else:
                        # This is very likely a timeout (RE Manager is not responding)
                        raise Exception("timeout occurred")
                except Exception as ex:
                    msg_err = f"ZeroMQ communication failed: {str(ex)}"
                    self._zmq_error_occurred = True

                if self._cb:
                    try:
                        self._cb(msg, msg_err)
                    except Exception as ex:
                        # Operation should continue even if there are problems with callback
                        logger.exception("Exception occurred in ZMQ communication callback: %s", str(ex))

                self._event_wait_for_msg.clear()
                if self._blocking_call:
                    self._event_msg_received.set()
                else:
                    # Release the lock here if the call is not blocking.
                    self._lock_zmq.release()

            if not self._thread_running:  # Exit thread
                break

    def _start_receive_thread(self):
        """
        Start the receive thread
        """
        # Initializations
        self._cb = None
        self._event_msg_received.clear()
        self._event_wait_for_msg.clear()

        # Start 'receive' thread
        if not self._thread_running:
            self._thread_running = True
            self._zmq_receive_thread = threading.Thread(
                target=self._receive_thread, name=self._thread_name, daemon=True
            )
            self._zmq_receive_thread.start()

    def _zmq_receive_blocking_cb(self, msg, msg_err):
        """
        Callback function used when ``send_message()`` is called in blocking mode.
        This function can also be used as an example of a user-defined callback function
        in case of non-blocking calls.

        Parameters
        ----------
        msg: dict
            Dictionary of parameters returned by the server.
        msg_err: str
            Error message: ``""`` if communication was successful, error message
            if communication error (most likely timeout) occurred.
        """
        self._received_data = {"msg": msg, "msg_err": msg_err}
        self._event_msg_received.set()

    def _zmq_communicate(self, msg_out, *, cb):
        """
        Send message to the server and receive the response.

        Parameters
        ----------
        msg_out: dict
            Dictionary that contains outgoing message (method name and parameters).
        cb: callable or None
            Reference to callback function. If callback function is specified, then
            the `_zmq_communicate` sends outgoing message and exits. Callback function
            is called once the response is received from the server or timeout occurs.

        Returns
        -------
        msg: dict or None
            Returns dictionary with parameters returned by the server or communication
            error message if the function is called in blocking mode (no callback is
            specified). In non-blocking mode the function always returns ``None``.
        """

        # Blocking attempt to acquire the lock.
        self._lock_zmq.acquire()

        # Initializations
        self._event_msg_received.clear()
        self._event_wait_for_msg.clear()
        self._received_data = {}

        # Restart the socket
        if self._zmq_error_occurred:
            self._zmq_socket_restart()
            self._zmq_error_occurred = False

        if cb is None:
            self._cb = self._zmq_receive_blocking_cb
            self._blocking_call = True
        else:
            self._cb = cb
            self._blocking_call = False

        self._zmq_socket.send_json(msg_out)
        self._event_wait_for_msg.set()

        if self._blocking_call:
            self._event_msg_received.wait()
            self._event_msg_received.clear()

            msg, msg_err = {}, ""
            if self._received_data:
                msg, msg_err = self._received_data["msg"], self._received_data["msg_err"]
            else:
                msg_err = "No received data was found"

            # Release the lock here if the call is blocking. If the call is not blocking
            #   then release the lock in the communication thread.
            self._lock_zmq.release()

            if msg_err:
                raise CommTimeoutError(msg_err)

            return msg
        else:
            # There is nothing to return if the call is not blocking.
            return None

    def _zmq_socket_open(self):
        """
        Open ZMQ socket.
        """
        self._zmq_socket = self._ctx.socket(zmq.REQ)

        if self._server_public_key:
            # Set server public key
            self._zmq_socket.set(zmq.CURVE_SERVERKEY, self._server_public_key.encode("utf-8"))
            # Set public and private keys for the client
            self._zmq_socket.set(zmq.CURVE_PUBLICKEY, _fixed_public_key.encode("utf-8"))
            self._zmq_socket.set(zmq.CURVE_SECRETKEY, _fixed_private_key.encode("utf-8"))

        # Increment `self._timeout_receive` so that timeout supplied to `self._zmq_socket.poll`
        #   expires first so that correct message is produced.
        self._zmq_socket.RCVTIMEO = self._timeout_receive + 1
        self._zmq_socket.SNDTIMEO = self._timeout_send
        # Clear the buffer quickly after the socket is closed
        self._zmq_socket.setsockopt(zmq.LINGER, 100)

        # Successful connection does not mean that the socket exists
        self._zmq_socket.connect(self._zmq_server_address)

        logger.info("Connected to ZeroMQ server '%s'" % str(self._zmq_server_address))
        logger.info("ZMQ encryption: %s", "disabled" if self._server_public_key is None else "enabled")

    def _zmq_socket_restart(self):
        """
        Restart (close and open the socket). Should be called in case of communication
        error (e.g. the server did not return the response).
        """
        self._zmq_socket.close()
        self._zmq_socket_open()

    def _create_msg(self, *, method, params=None):
        """
        Format the message.

        Parameters
        ----------
        method: str
            Method name
        params: dict or None
            Dictionary of parameters or None if no parameters should be supplied for the method.
        """
        return {"method": method, "params": params}

    def send_message(self, *, method, params=None, cb=None, raise_exceptions=None):
        """
        Send message to ZMQ server and wait for the response. The message must contain
        a name of a method supported by the server and a dictionary of parameters that
        are required by the method. In case of communication error (timeout), the function
        returns error message or raises ``CommTimeoutError`` exception depending on
        the values ``raise_exceptions`` parameter in this function and class constructor.

        Parameters
        ----------
        method: str
            Name of the method to be invoked on the server. The method must be supported
            by the server.
        params: dict or None
            Dictionary of parameters passed to the method. If ``None``, then an empty dictionary
            is passed to the server.
        cb: callable or None
            Callback function. If None, then the function blocks until communication is complete.
            Otherwise the function exits after the message is sent to the server. The callback
            is called when the response is received or timeout occurs.
            Function signature is ``cb(msg: dict, msg_err: str)``. Here ``msg`` is the dictionary
            of the parameters returned by the server and ``msg_err`` is ``""`` if communication
            was successful and contains error message in case communication failed (timeout).
        raise_exceptions: bool or None
            The flag indicates if exception should be raised in case of communication error
            (such as timeout). If ``False``, then error message is returned instead.
            If None, then the field value ``self._raise_timeout_exceptions`` is used to
            determine if the exception needs to be raised. Non-blocking calls do not raise
            the exception. Instead, callback function receives the error message as one of the
            parameters

        Returns
        -------
        dict
            Message returned by the server.

        Raises
        ------
        CommTimeoutError
            Raised if communication error occurs and ``raise_exceptions`` is set ``True``.
        """

        # Send empty dictionary if no parameters are passed
        params = params or {}

        try:
            msg_out = self._create_msg(method=method, params=params)
            msg_in = self._zmq_communicate(msg_out, cb=cb)
        except Exception as ex:
            errmsg = f"ZMQ communication error: {str(ex)}"
            use_ex = raise_exceptions if (raise_exceptions is not None) else self._raise_exceptions
            if use_ex:
                raise CommTimeoutError(errmsg)
            msg_in = {"success": False, "msg": errmsg}

        return msg_in

    def close(self):
        """
        Close ZMQ socket. Call to close socket if the object is no longer needed, but may
        not be destroyed for some time.
        """
        self._zmq_socket.close()
        self._thread_running = False


class ZMQCommSendAsync:
    """
    API for communication with RE Manager via ZMQ. The object has to be created
    from the running even loop or the loop has to be passed as a parameter during
    initialization.

    Parameters
    ----------
    loop : asyncio loop
        Current event loop
    zmq_server_address : str or None
        Address of ZMQ server. If None, then the default address is ``tcp://localhost:60615``
        is used.
    timeout_recv : int
        Timeout (in ms) for ZMQ receive operations.
    timeout_send: int
        Timeout (in ms) for ZMQ send operations.
    raise_exceptions : bool
        Tells if exceptions should be raised in case of communication errors (mostly timeouts)
        when ``send_message()`` is awaited. This setting can be overridden by
        specifying ``raise_exceptions`` parameter of ``send_message()`` function. The exception
        ``CommTimeoutError`` is raised if the parameter is ``True``, otherwise error message
        is returned by ``send_message()``.
    server_public_key : str or None
        Server public key (z85-encoded 40 character string). The valid public key from the server
        public/private key pair must be passed if encryption is enabled at the 0MQ server side.
        Communication requests will time out if the key is invalid. Exception will be raised if
        the key is improperly formatted. Encryption will be disabled if ``None`` is passed.

    Examples
    --------

    .. code-block:: python

        async def communicate():
            zmq_comm = ZMQCommSendAsync()
            for n in range(10):
                msg = await zmq_comm.send_message(method="some_method", params={"some_value": n})
                print(f"msg={msg}")
            zmq_comm.close()

        asyncio.run(communicate())

    """

    def __init__(
        self,
        *,
        loop=None,
        zmq_server_address=None,
        timeout_recv=2000,
        timeout_send=500,
        raise_exceptions=True,
        server_public_key=None,
    ):
        self._loop = loop if loop else asyncio.get_event_loop()

        zmq_server_address = zmq_server_address or "tcp://localhost:60615"
        self._server_public_key = server_public_key

        self._timeout_receive = timeout_recv  # Timeout for 'recv' operation (ms)
        self._timeout_send = timeout_send  # # Timeout for 'send' operation (ms)
        self._raise_exceptions = raise_exceptions

        # ZeroMQ communication
        self._ctx = zmq.asyncio.Context()
        self._zmq_socket = None
        self._zmq_server_address = zmq_server_address

        if self._server_public_key is not None:
            validate_zmq_key(self._server_public_key)

        self._zmq_socket_open()
        self._lock_zmq = asyncio.Lock()

    def __del__(self):
        self.close()

    def get_loop(self):
        """
        Returns the asyncio event loop.
        """
        return self._loop

    async def _zmq_send(self, msg):
        await self._zmq_socket.send_json(msg)

    async def _zmq_receive(self):
        try:
            if await self._zmq_socket.poll(timeout=self._timeout_receive):
                msg = await self._zmq_socket.recv_json()
            else:
                # This is very likely a timeout (RE Manager is not responding)
                raise Exception("timeout occurred")
        except Exception as ex:
            # Timeout occurred. Socket needs to be reset.
            logger.exception("ZeroMQ communication failed: %s" % str(ex))
            raise
        return msg

    async def _zmq_communicate(self, msg_out):
        await self._zmq_send(msg_out)
        msg_in = await self._zmq_receive()
        return msg_in

    def _zmq_socket_open(self):
        self._zmq_socket = self._ctx.socket(zmq.REQ)

        if self._server_public_key:
            # Set server public key
            self._zmq_socket.set(zmq.CURVE_SERVERKEY, self._server_public_key.encode("utf-8"))
            # Set public and private keys for the client
            self._zmq_socket.set(zmq.CURVE_PUBLICKEY, _fixed_public_key.encode("utf-8"))
            self._zmq_socket.set(zmq.CURVE_SECRETKEY, _fixed_private_key.encode("utf-8"))

        # Increment `self._timeout_receive` so that timeout supplied to `self._zmq_socket.poll`
        #   expires first so that correct message is produced.
        self._zmq_socket.RCVTIMEO = self._timeout_receive + 1
        self._zmq_socket.SNDTIMEO = self._timeout_send
        # Clear the buffer quickly after the socket is closed
        self._zmq_socket.setsockopt(zmq.LINGER, 100)

        # Successful connection does not mean that the socket exists
        self._zmq_socket.connect(self._zmq_server_address)

        logger.info("Connected to ZeroMQ server '%s'" % str(self._zmq_server_address))
        logger.info("ZMQ encryption: %s", "disabled" if self._server_public_key is None else "enabled")

    def _zmq_socket_restart(self):
        self._zmq_socket.close()
        self._zmq_socket_open()

    def _create_msg(self, *, method, params=None):
        return {"method": method, "params": params}

    async def send_message(self, *, method, params=None, raise_exceptions=None):
        """
        Send message to ZMQ server and wait for the response. The message must contain
        a name of a method supported by the server and a dictionary of parameters that
        are required by the method. In case of communication error (timeout), the function
        returns error message or raises ``CommTimeoutError`` exception depending on
        the values ``raise_exceptions`` parameter in this function and class constructor.

        Parameters
        ----------
        method: str
            Name of the method to be invoked on the server. The method must be supported
            by the server.
        params: dict or None
            Dictionary of parameters passed to the method. If ``None``, then an empty dictionary
            is passed to the server.
        raise_exceptions: bool or None
            The flag indicates if exception should be raised in case of communication error
            (such as timeout). If ``False``, then error message is returned instead.
            If None, then the field value ``self._raise_exceptions`` is used to
            determine if the exception needs to be raised. Non-blocking calls do not raise
            the exception. Instead, callback function receives the error message as one of the
            parameters

        Returns
        -------
        dict
            Message returned by the server.

        Raises
        ------
        CommTimeoutError
            Raised if communication error occurs and ``raise_timeout_exceptions`` is set ``True``.
        """

        # Send empty dictionary if no parameters are passed
        params = params or {}

        async with self._lock_zmq:
            try:
                msg_out = self._create_msg(method=method, params=params)
                msg_in = await self._zmq_communicate(msg_out)
            except Exception as ex:
                # This is very likely a timeout (RE Manager is not responding)
                self._zmq_socket_restart()
                errmsg = f"ZMQ communication error: {str(ex)}"
                use_ex = raise_exceptions if (raise_exceptions is not None) else self._raise_exceptions
                if use_ex:
                    raise CommTimeoutError(errmsg)
                msg_in = {"success": False, "msg": errmsg}
            return msg_in

    def close(self):
        """
        Close ZMQ socket. Call to close socket if the object is no longer needed, but may
        not be destroyed for some time.
        """
        if self._zmq_socket:
            self._zmq_socket.close()


def zmq_single_request(method, params=None, *, zmq_server_address=None, server_public_key=None):
    """
    Send a single request to ZMQ server. The function opens the socket, sends
    a single ZMQ request and closes the socket. The function is not expected
    to raise exceptions. In case of communication error the return value
    of ``msg`` is ``None`` and ``err_msg`` contains the error message. Otherwise
    ``err_msg`` is empty and ``msg`` contains the dictionary returned by the server.

    Parameters
    ----------
    method: str
        Name of the method called in RE Manager
    params: dict or None
        Dictionary of parameters (payload of the message). If ``None`` then
        the message is sent with empty payload: ``params = {}``.
    server_public_key: str or None
        Server public key (z85-encoded 40 character string). The Valid public key from the server
        public/private key pair must be passed if encryption is enabled at the 0MQ server side.
        Communication requests will time out if the key is invalid. Exception will be raised if
        the key is improperly formatted. Encryption will be disabled if ``None`` is passed.


    Returns
    -------
    msg: dict or None
        Message received from RE Manager in response to the request. None if communication
        error (timeout) occurred.
    err_msg: str
        Contains a message in case communication error (timeout) occurs. Empty string otherwise.
    """

    msg_received = None

    async def send_request(method, params):

        nonlocal msg_received
        zmq_to_manager = ZMQCommSendAsync(
            zmq_server_address=zmq_server_address, server_public_key=server_public_key
        )
        msg_received = await zmq_to_manager.send_message(method=method, params=params, raise_exceptions=True)
        zmq_to_manager.close()

    try:
        asyncio.run(send_request(method, params))

        msg = msg_received
        msg_err = ""
    except Exception as ex:
        msg = None
        msg_err = str(ex)

    if msg_err:
        logger.warning("Communication with RE Manager failed: %s", str(msg_err))

    return msg, msg_err



# flake8: noqa
from typing import List, Optional, Dict, Any

from bluesky.plans import (
    count,
    list_scan,
    rel_list_scan,
    list_grid_scan,
    rel_list_grid_scan,
    log_scan,
    rel_log_scan,
    adaptive_scan,
    rel_adaptive_scan,
    tune_centroid,
    scan_nd,
    inner_product_scan,
    scan,
    grid_scan,
    rel_grid_scan,
    relative_inner_product_scan,
    rel_scan,
    tweak,
    spiral_fermat,
    rel_spiral_fermat,
    spiral,
    rel_spiral,
    spiral_square,
    rel_spiral_square,
    ramp_plan,
    fly,
    x2x_scan,
)


def marked_up_count(
    detectors: List[Any], num: int = 1, delay: Optional[float] = None, md: Optional[Dict[str, Any]] = None
):
    return (yield from count(detectors, num=num, delay=delay, md=md))



import os
import glob
import shutil
import pytest
import subprocess
import asyncio
import time as ttime
import intake
import tempfile
import sys

from databroker import catalog_search_path

from bluesky_queueserver.manager.profile_ops import get_default_startup_dir
from bluesky_queueserver.manager.plan_queue_ops import PlanQueueOperations
from bluesky_queueserver.manager.comms import zmq_single_request

import logging

logger = logging.Logger(__name__)


def copy_default_profile_collection(tmp_path, *, copy_py=True, copy_yaml=True):
    """
    Copy default profile collections (only .py files) to temporary directory.
    Returns the new temporary directory.
    """
    # Default path
    pc_path = get_default_startup_dir()
    # New path
    new_pc_path = os.path.join(tmp_path, "startup")

    os.makedirs(new_pc_path, exist_ok=True)

    # Copy simulated profile collection (only .py files)
    patterns = []
    if copy_py:
        patterns.extend(["*.py", "*.ipy"])
    if copy_yaml:
        patterns.append("*.yaml")
    for pattern in patterns:
        file_pattern = os.path.join(pc_path, pattern)
        file_list = glob.glob(file_pattern)
        for fln in file_list:
            shutil.copy(fln, new_pc_path)

    return new_pc_path


def patch_first_startup_file(pc_path, additional_code):
    """
    Adds code to the beginning of a startup file.

    Parameters
    ----------
    pc_path: str
        Path to the directory with profile collection
    additional_code: str
        Code (text)that should be added to the beginning of the startup file
    """

    # Path to the first file (starts with 00)
    file_pattern_py = os.path.join(pc_path, "*.py")
    file_pattern_ipy = os.path.join(pc_path, "*.ipy")
    file_list = glob.glob(file_pattern_py) + glob.glob(file_pattern_ipy)
    file_list.sort()
    fln = file_list[0]

    # Backup the file which is about to be changed if it was not backed up already.
    fln_tmp = os.path.join(pc_path, "_backup")
    if not os.path.exists(fln_tmp):
        shutil.copy(fln, fln_tmp)

    with open(fln, "r") as file_in:
        code = file_in.readlines()

    with open(fln, "w") as file_out:
        file_out.writelines(additional_code)
        file_out.writelines(code)


def patch_first_startup_file_undo(pc_path):
    """
    Remove patches applied to the first file of profile collection.
    """
    # Path to the first file (starts with 00)
    file_pattern_py = os.path.join(pc_path, "*.py")
    file_pattern_ipy = os.path.join(pc_path, "*.ipy")
    file_list = glob.glob(file_pattern_py) + glob.glob(file_pattern_ipy)
    file_list.sort()
    fln = file_list[0]

    fln_tmp = os.path.join(pc_path, "_backup")
    if os.path.isfile(fln_tmp):
        os.remove(fln)
        os.rename(fln_tmp, fln)


def append_code_to_last_startup_file(pc_path, additional_code):
    """
    Append code to the end of the last startup file.

    Parameters
    ----------
    pc_path: str
        Path to the directory with IPython ``startup`` directory
    additional_code: str
        Code (text) that should be added to the end of the startup file
    """

    # Path to the last file
    file_pattern_py = os.path.join(pc_path, "*.py")
    file_pattern_ipy = os.path.join(pc_path, "*.ipy")
    file_list = glob.glob(file_pattern_py) + glob.glob(file_pattern_ipy)
    file_list.sort()
    fln = file_list[-1]

    # Backup the file which is about to be changed if it was not backed up already.
    fln_tmp = os.path.join(pc_path, "_backup")
    if not os.path.exists(fln_tmp):
        shutil.copy(fln, fln_tmp)

    with open(fln, "r") as file_in:
        code = file_in.readlines()

    with open(fln, "w") as file_out:
        file_out.writelines(code)
        file_out.write("\n\n")
        file_out.writelines(additional_code)


# The name of env. variable holding ZMQ public key. Public key is necessary in tests using encryption
#   and passing it using env variable (set using monkeypatch) is convenient.
_name_ev_public_key = "_TEST_QSERVER_ZMQ_PUBLIC_KEY_"


def set_qserver_zmq_public_key(mpatch, *, server_public_key):
    """
    Temporarily set environment variable holding server public key for using ``zmq_secure_request``.
    Note, that ``monkeypatch`` should precede the fixture that is starting RE Manager in test
    parameters. Otherwise the environment variable will be cleared before RE Manager is stopped and
    the test will fail.

    Parameters
    ----------
    mpatch
        instance of ``monkeypatch``.
    server_public_key : str
        ZMQ server public key represented as 40 character z85 encoded string.
    """
    mpatch.setenv(_name_ev_public_key, server_public_key)


def clear_qserver_zmq_public_key(mpatch):
    """
    Clear the environment variable holding server public key set by ``set_qserver_zmq_public_key``.
    """
    mpatch.delenv(_name_ev_public_key)


_name_ev_zmq_address = "_TEST_QSERVER_ZMQ_ADDRESS_"


def set_qserver_zmq_address(mpatch, *, zmq_server_address):
    """
    Temporarily set environment variable holding server zmq address for using with ``zmq_secure_request``.
    Note, that ``monkeypatch`` should precede the fixture that is starting RE Manager in test
    parameters. Otherwise the environment variable will be cleared before RE Manager is stopped and
    the test will fail.

    Parameters
    ----------
    mpatch
        instance of ``monkeypatch``.
    server_zmq_address : str
        ZMQ address (such as 'tcp://localhost:60615').
    """
    mpatch.setenv(_name_ev_zmq_address, zmq_server_address)


def clear_qserver_zmq_address(mpatch):
    """
    Clear the environment variable holding server zmq address set by ``set_qserver_zmq_address``.
    """
    mpatch.delenv(_name_ev_zmq_address)


def zmq_secure_request(method, params=None, *, zmq_server_address=None):
    """
    Wrapper for 'zmq_single_request'. Verifies if environment variable holding server public key is set
    and passes the key to 'zmq_single_request' . Simplifies writing tests that use RE Manager in secure mode.
    Use functions `set_qserver_zmq_public_key()` and `clear_qserver_zmq_public_key()` to set and
    clear the environment variable.

    The function also verifies if the environment variable holding ZMQ server address is set, and
    passes the address to ``zmq_single_request``. If ``zmq_server_address`` is passed as a parameter, then
    the environment variable is ignored (at least in current implementation).
    """
    server_public_key = None

    if _name_ev_public_key in os.environ:
        server_public_key = os.environ[_name_ev_public_key]

    # Use the address passed in environment variable only if the parameter 'zmq_server_address' is None
    if (_name_ev_zmq_address in os.environ) and (zmq_server_address is None):
        zmq_server_address = os.environ[_name_ev_zmq_address]

    return zmq_single_request(
        method=method, params=params, zmq_server_address=zmq_server_address, server_public_key=server_public_key
    )


def get_queue_state():
    method, params = "status", None
    msg, _ = zmq_secure_request(method, params)
    if msg is None:
        raise TimeoutError("Timeout occurred while reading RE Manager status.")
    return msg


def get_queue():
    """
    Returns current queue.
    """
    method, params = "queue_get", None
    msg, _ = zmq_secure_request(method, params)
    if msg is None:
        raise TimeoutError("Timeout occurred while loading queue from RE Manager.")
    return msg


def get_reduced_state_info():
    msg = get_queue_state()
    items_in_queue = msg["items_in_queue"]
    queue_is_running = msg["manager_state"] == "executing_queue"
    items_in_history = msg["items_in_history"]
    return items_in_queue, queue_is_running, items_in_history


def condition_manager_idle(msg):
    return ("manager_state" in msg) and (msg["manager_state"] == "idle")


def condition_manager_paused(msg):
    return msg["manager_state"] == "paused"


def condition_environment_created(msg):
    return msg["worker_environment_exists"] and (msg["manager_state"] == "idle")


def condition_environment_closed(msg):
    return (not msg["worker_environment_exists"]) and (msg["manager_state"] == "idle")


def condition_queue_processing_finished(msg):
    items_in_queue = msg["items_in_queue"]
    queue_is_running = msg["manager_state"] == "executing_queue"
    return (items_in_queue == 0) and not queue_is_running


def wait_for_condition(time, condition):
    """
    Wait until queue is processed. Note: processing of TimeoutError is needed for
    monitoring RE Manager while it is restarted.
    """
    dt = 0.5  # Period for checking the queue status
    time_stop = ttime.time() + time
    while ttime.time() < time_stop:
        ttime.sleep(dt / 2)
        try:
            msg = get_queue_state()
            if condition(msg):
                return True
        except TimeoutError:
            pass
        ttime.sleep(dt / 2)
    return False


def wait_for_task_result(time, task_uid):
    """
    Wait for the results of the task defined by ``task_uid``. Raises ``TimeoutError`` if
    timeout ``time`` is exceeded while waiting for the task result. Returns the task result
    of the task only if it was completed.
    """
    dt = 0.2  # Period for checking if the task is available
    time_stop = ttime.time() + time
    while ttime.time() < time_stop:
        ttime.sleep(dt / 2)
        try:
            resp, _ = zmq_secure_request("task_result", params={"task_uid": task_uid})

            assert resp["success"] is True, f"Request for task result failed: {resp['msg']}"
            assert resp["task_uid"] == task_uid

            if resp["status"] == "completed":
                return resp["result"]

        except TimeoutError:
            pass
        ttime.sleep(dt / 2)

    raise TimeoutError(f"Timeout occurred while waiting for results of the task {task_uid!r}")


def clear_redis_pool():
    # Remove all Redis entries.
    pq = PlanQueueOperations()

    async def run():
        await pq.start()
        await pq.delete_pool_entries()
        await pq.user_group_permissions_clear()

    asyncio.run(run())


@pytest.fixture(scope="module")
def db_catalog():
    """
    Creates msgpack-based catalog, returns reference to the catalog and the catalog name.
    The catalog name may be used as a configuration name for subscribing to databroker.
    Yields the dictionary: ``db_catalog["catalog"]`` - reference to the catalog,
    ``db_catalog["catalog_name"]`` - string that represents the catalog name.
    """
    db_catalog_name = "qserver_tests"

    config_dir = catalog_search_path()[0]
    config_path = os.path.join(config_dir, f"{db_catalog_name}.yml")

    files_dir = os.path.join(tempfile.gettempdir(), "qserver_tests", "db_catalog_files")
    files_dir = os.path.abspath(files_dir)
    files_path = os.path.join(files_dir, "*.msgpack")

    # Delete the directory 'db_catalog_files' and its contents in case it exists.
    if os.path.isdir(files_dir):
        shutil.rmtree(files_dir)

    os.makedirs(config_dir, exist_ok=True)
    os.makedirs(files_dir, exist_ok=True)

    config_file_contents = f"""
sources:
  {db_catalog_name}:
    driver: bluesky-msgpack-catalog
    args:
      paths:
        - "{files_path}"
"""

    with open(config_path, "w") as file_out:
        file_out.writelines(config_file_contents)

    # The catalog can not be opened using intake right away:
    #   cat = intake.open_catalog(config_path)
    # But standard way of opening a catalog such as
    #   from databroker import catalog
    #   catalog[qserver_tests]
    # and subscription of Run Engine to Data Broker will not work.
    # So we need the delay.
    ttime.sleep(2.0)

    cat = intake.open_catalog(config_path)
    cat = cat[db_catalog_name]

    yield {"catalog": cat, "catalog_name": db_catalog_name}

    os.remove(config_path)
    shutil.rmtree(files_dir)


class ReManager:
    def __init__(self, params=None, *, stdout=sys.stdout, stderr=sys.stdout):
        self._p = None
        self.start_manager(params, stdout=stdout, stderr=stderr)

    def start_manager(self, params=None, *, stdout=sys.stdout, stderr=sys.stdout, cleanup=True):
        """
        Start RE manager.

        Parameters
        ----------
        params : list(str)
            The list of command line parameters passed to the manager at startup
        stdout
            Device to forward stdout, ``None`` - print to console
        stderr
            Device to forward stdout, ``None`` - print to console
        cleanup: boolean
            Perform cleanup (remove related Redis keys) before opening the manager.

        Returns
        -------
        subprocess.Popen
            Subprocess in which the manager is running. It needs to be stopped at certain point.
        """
        params = params or []

        # Set logging level for RE Manager unless it is already set in parameters
        #   Leads to excessive output and some tests may fail at tearup. Enable only when needed.
        logging_levels = ("--verbose", "--quiet", "--silent")
        if not any([_ in params for _ in logging_levels]):
            params.append("--verbose")

        if not self._p:
            if cleanup:
                clear_redis_pool()

            cmd = ["start-re-manager"]
            if params:
                cmd += params
            self._p = subprocess.Popen(cmd, universal_newlines=True, stdout=stdout, stderr=stderr)

    def check_if_stopped(self, timeout=5):
        """
        Waits for the process to be stopped. Returns ``True`` when it is stopped.
        and ``False`` otherwise.
        """
        try:
            # Make sure that the process is terminated
            self._p.wait(5)
            # Disable additional attempts to stop the process
            self._p = None
            return True
        except subprocess.TimeoutExpired:
            return False

    def stop_manager(self, timeout=10, cleanup=True):
        """
        Attempt to exit the subprocess that is running manager in orderly way and kill it
        after timeout.

        Parameters
        ----------
        timeout: float
            Timeout in seconds.
        cleanup: boolean
            Perform cleanup (remove created Redis keys) after closing the manager.
        """
        if self._p:
            try:
                # If the process is already terminated, then don't attempt to communicate with it.
                if self._p.poll() is None:
                    success, msg = False, ""

                    # Try to stop the manager in a nice way first by sending the command
                    resp1, err_msg1 = zmq_secure_request(method="manager_stop", params=None)
                    assert resp1, str(err_msg1)
                    success = resp1["success"]

                    if not success:
                        # If the command was rejected, try 'safe_off' mode that kills the RE worker environment
                        msg += f"Request to stop the manager in with the 'safe_on' option failed: {resp1['msg']}."
                        resp2, err_msg2 = zmq_secure_request(method="manager_stop", params={"option": "safe_off"})
                        assert resp2, str(err_msg2)
                        success = resp2["success"]
                        if not success:
                            msg += (
                                " Request to stop the manager in with the 'safe_off' "
                                f"option failed: {resp2['msg']}."
                            )
                        else:
                            msg += " Attempting to stop the manager using 'safe_off' option"

                    # If both requests failed, then there is nothing to wait
                    if not success:
                        raise RuntimeError(msg)

                    # If only the first request failed, then there is a chance that the manager stops
                    try:
                        self._p.communicate(timeout=timeout)
                        self._p = None
                    except subprocess.TimeoutExpired:
                        raise RuntimeError(f"Timeout occured while waiting for the manager to stop: {msg}")

                    # If at least one of the requests failed and there is an error message, then fail the test
                    if msg:
                        raise RuntimeError(msg)

            except Exception as ex:
                # The manager is not responsive, so just kill the process.
                if self._p:
                    self._p.kill()
                    self._p.wait(timeout)
                assert False, f"RE Manager failed to stop: {str(ex)}"

            finally:
                if cleanup:
                    clear_redis_pool()
                self._p = None

    def kill_manager(self, timeout=10):
        """
        Kill the subprocess running RE Manager. Use only to stop RE Manager that failed to start correctly.
        The operation will always succeed.

        Parameters
        ----------
        timeout: float
            Timeout in seconds.
        """
        if self._p:
            self._p.kill()
            self._p.wait(timeout)
            clear_redis_pool()


def _reset_queue_mode():
    """Reset queue mode to the default mode"""
    resp, msg = zmq_secure_request("queue_mode_set", params={"mode": "default"})
    if resp["success"] is not True:
        raise RuntimeError(msg)


@pytest.fixture
def re_manager_cmd():
    """
    Start RE Manager by running `start-re-manager` as a subprocess. Pass the list of
    command-line parameters to `start-re-manager`. Tests will communicate with RE Manager via ZeroMQ.

    Examples of using the fixture:
    ``re_manager_cmd()`` or ``re_manager_cmd([]) - create RE Manager without command-line parameters
    ``re_manager_cmd(["-h"])`` - equivalent to ``start-re-manager -h``.
    """
    re = {"re": None}
    failed_to_start = False

    def _create(params, *, stdout, stderr):
        """
        Create RE Manager. ``start-re-manager`` is called with command line parameters from
          the list ``params``.
        """
        nonlocal re, failed_to_start
        re["re"] = ReManager(params, stdout=stdout, stderr=stderr)

        # Wait until RE Manager is started. Raise exception if the server failed to start.
        if not wait_for_condition(time=10, condition=condition_manager_idle):
            failed_to_start = True
            raise TimeoutError("Timeout: RE Manager failed to start.")

        _reset_queue_mode()

    def _close():
        """
        Close RE Manager if it exists.
        """
        nonlocal re, failed_to_start
        if re["re"] is not None:
            if not failed_to_start:
                re["re"].stop_manager()
            else:
                re["re"].kill_manager()

    def create_re_manager(params=None, *, stdout=sys.stdout, stderr=sys.stdout):
        params = params or []
        _close()
        _create(params, stdout=stdout, stderr=stderr)
        return re["re"]

    yield create_re_manager

    _close()


@pytest.fixture
def re_manager():
    """
    Start RE Manager as a subprocess. Tests will communicate with RE Manager via ZeroMQ.
    """
    re = ReManager()
    failed_to_start = False

    # Wait until RE Manager is started. Raise exception if the server failed to start.
    if not wait_for_condition(time=10, condition=condition_manager_idle):
        failed_to_start = True
        raise TimeoutError("Timeout: RE Manager failed to start.")

    _reset_queue_mode()

    yield re  # Nothing to return
    if not failed_to_start:
        re.stop_manager()
    else:
        re.kill_manager()


@pytest.fixture
def re_manager_pc_copy(tmp_path):
    """
    Start RE Manager as a subprocess. Tests will communicate with RE Manager via ZeroMQ.
    Copy profile collection and return its temporary path.
    """
    pc_path = copy_default_profile_collection(tmp_path)
    re = ReManager(["--startup-dir", pc_path])
    failed_to_start = False

    # Wait until RE Manager is started. Raise exception if the server failed to start.
    if not wait_for_condition(time=10, condition=condition_manager_idle):
        failed_to_start = True
        raise TimeoutError("Timeout: RE Manager failed to start.")

    _reset_queue_mode()

    yield re, pc_path  # Location of the copy of the default profile collection.
    if not failed_to_start:
        re.stop_manager()
    else:
        re.kill_manager()


@pytest.fixture
def reset_sys_modules():
    """
    Resets `sys.modules` after each test. Allows importing the same module in multiple tests independently.
    Intended for use in the tests for the functions that import modules or execute scripts that import modules.
    """
    # Save the set of keys before the test is called
    sys_modules = list(sys.modules.keys())

    yield

    # Remove entries for all the modules loaded during the test
    for key in list(sys.modules.keys()):
        if key not in sys_modules:
            print(f"Deleting the key '{key}'")
            del sys.modules[key]
